# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020, Yoel Cortes-Pena, <yoelcortes@gmail.com>, Yalin Li, <mailto.yalin.li@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
import thermosteam as tmo
from thermosteam import functional as fn
import flexsolve as flx
from chemicals import elements
from warnings import warn
from ..base import SparseVector, SparseArray, DictionaryView
from . import (
    _parse as prs,
    _xparse as xprs,
)
from ..utils import chemicals_user
from .._phase import NoPhase, phase_tuple
from ..indexer import ChemicalIndexer, MaterialIndexer
from ..exceptions import InfeasibleRegion
import numpy as np
import pandas as pd

__all__ = ('Reaction', 'ReactionItem', 'ReactionSet', 
           'ParallelReaction', 'SeriesReaction', 'ReactionSystem',
           'Rxn', 'RxnS', 'RxnI', 'PRxn', 'SRxn', 'RxnSys')

def get_stoichiometric_string(stoichiometry, phases, chemicals):
    if phases:
        return xprs.get_stoichiometric_string(stoichiometry,
                                              phases,
                                              chemicals)
    else:
        return prs.get_stoichiometric_string(stoichiometry, 
                                             chemicals)

def set_reaction_basis(rxn, basis):
    if basis != rxn._basis:
        if basis == 'wt':
            if isinstance(rxn, ReactionSet):
                for i in rxn._stoichiometry: i *= rxn.MWs
            else:
                rxn._stoichiometry *= rxn.MWs
        elif basis == 'mol':
            if isinstance(rxn, ReactionSet):
                for i in rxn._stoichiometry: i /= rxn.MWs
            else:
                rxn._stoichiometry /= rxn.MWs
        else:
            raise ValueError("basis must be either by 'wt' or by 'mol'")
        rxn._rescale()
        rxn._basis = basis

def as_material_array(material, basis, phases, chemicals):
    """
    Returns the material flows as a sparse array (or vector),
    the old configuration if chemicals changed, whether the sparse data dictionary
    is wrapped, and whether material flows is a copy.
    
    """
    isa = isinstance
    if isa(material, tmo.Stream):
        if phases and material.phases != phases:
            raise ValueError("reaction and stream phases do not match")
        if material.chemicals is chemicals:
            config = None
        else:
            config = material.chemicals, material.imol.reset_chemicals(chemicals)
        if basis == 'mol':
            return material._imol.data, config, None
        elif basis == 'wt':
            original = material.imass.data
            return original.copy(), config, original
        else:
            raise ValueError("basis must be either 'mol' or 'wt'")
    elif material.__class__ in (SparseVector, SparseArray):
        ndim = material.ndim
        if ndim == 1:
            if isinstance(material.dct, DictionaryView):  
                return (material.copy(), None, material)
            else:
                return (material, None, None)
        elif ndim == 2:
            for i in material.rows:
                if isinstance(i.dct, DictionaryView):
                    return material.copy(), None, material
            else:
                return material.copy(), None, None
        else:
            raise Exception('unknown error')
    elif phases:
        return SparseArray(material), None, material
    else:
        return SparseVector(material), None, material


# %%

@chemicals_user
class Reaction:
    """
    Create a Reaction object which defines a stoichiometric reaction and
    conversion. A Reaction object is capable of reacting the material
    flow rates of a :class:`thermosteam.Stream` object.
    
    Parameters
    ----------
    reaction : dict or str
               A dictionary of stoichiometric coefficients or a stoichiometric
               equation written as:
               i1 R1 + ... + in Rn -> j1 P1 + ... + jm Pm
    reactant : str, optional
               ID of reactant chemical. Defaults to reactant if only one available.
    X : float, optional
        Reactant conversion (fraction). Defaults to 1.
    chemicals : Chemicals, optional
        Chemicals corresponding to each entry in the stoichiometry array. 
        Defaults to settings.chemicals
    basis: {'mol', 'wt'}, optional
        Basis of reaction. Defaults to 'mol'.
    
    Other Parameters
    ----------------
    check_mass_balance=False: bool
        Whether to check if mass is not created or destroyed.
    correct_mass_balance=False: bool
        Whether to make sure mass is not created or destroyed by varying the 
        reactant stoichiometric coefficient.
    check_atomic_balance=False: bool
        Whether to check if stoichiometric balance by atoms cancel out.
    correct_atomic_balance=False: bool
        Whether to correct the stoichiometry according to the atomic balance.
    
    Notes
    -----
    A reaction object can react either a stream or an array. When a stream
    is passed, it reacts either the mol or mass flow rate according to
    the basis of the reaction object. When an array is passed, the array
    elements are reacted regardless of what basis they are associated with.
    
    Warning
    -------
    Negative conversions and conversions above 1.0 are fair game (allowed), but
    may lead to odd/infeasible values when reacting a stream.
    
    Examples
    --------
    Electrolysis of water to molecular hydrogen and oxygen:
    
    >>> import thermosteam as tmo
    >>> chemicals = tmo.Chemicals(['H2O', 'H2', 'O2'], cache=True)
    >>> tmo.settings.set_thermo(chemicals)
    >>> reaction = tmo.Reaction('2H2O,l -> 2H2,g + O2,g', reactant='H2O', X=0.7)
    >>> reaction.show() # Note that the default basis is by 'mol'
    Reaction (by mol):
    stoichiometry             reactant    X[%]
    H2O,l -> H2,g + 0.5 O2,g  H2O,l      70.00
    >>> reaction.reactant # The reactant is a tuple of phase and chemical ID
    ('l', 'H2O')
    >>> feed = tmo.Stream('feed', H2O=100)
    >>> feed.phases = ('g', 'l') # Gas and liquid phases must be available
    >>> reaction(feed) # Call to run reaction on molar flow
    >>> feed.show() # Notice how 70% of water was converted to product
    MultiStream: feed
    phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
    flow (kmol/hr): (g) H2   70
                        O2   35
                    (l) H2O  30
    
    Let's change to a per 'wt' basis:
    
    >>> reaction.basis = 'wt'
    >>> reaction.show()
    Reaction (by wt):
    stoichiometry                     reactant    X[%]
    H2O,l -> 0.112 H2,g + 0.888 O2,g  H2O,l      70.00
    
    Although we changed the basis, the end result is the same if we pass a 
    stream:
    
    >>> feed = tmo.Stream('feed', H2O=100)
    >>> feed.phases = ('g', 'l')
    >>> reaction(feed) # Call to run reaction on mass flow
    >>> feed.show() # Notice how 70% of water was converted to product
    MultiStream: feed
    phases: ('g', 'l'), T: 298.15 K, P: 101325 Pa
    flow (kmol/hr): (g) H2   70
                        O2   35
                    (l) H2O  30
    
    If chemicals phases are not specified, Reaction objects can 
    react a any single phase Stream object (regardless of phase):
    
    >>> reaction = tmo.Reaction('2H2O -> 2H2 + O2', reactant='H2O', X=0.7)
    >>> feed = tmo.Stream('feed', H2O=100, phase='g')
    >>> reaction(feed) 
    >>> feed.show() 
    Stream: feed
    phase: 'g', T: 298.15 K, P: 101325 Pa
    flow (kmol/hr): H2O  30
                    H2   70
                    O2   35
    
    Alternatively, it's also possible to react an array (instead of a stream):
    
    >>> import numpy as np
    >>> array = np.array([100., 0. , 0.])
    >>> reaction(array)
    >>> array
    array([30., 70., 35.])
    
    Reaction objects with the same reactant can be added together:
        
    >>> tmo.settings.set_thermo(['Glucose', 'Ethanol', 'H2O', 'O2', 'CO2'])
    >>> fermentation = tmo.Reaction('Glucose + O2 -> Ethanol + CO2', reactant='Glucose', X=0.7)
    >>> combustion = tmo.Reaction('Glucose + O2 -> H2O + CO2', reactant='Glucose', X=0.2)
    >>> mixed_reaction = fermentation + combustion
    >>> mixed_reaction.show()
    Reaction (by mol):
    stoichiometry                                    reactant    X[%]
    Glucose + O2 -> 0.778 Ethanol + 0.222 H2O + CO2  Glucose    90.00
    
    Note how conversions are added and the stoichiometry rescales to a per
    reactant basis. Conversely, reaction objects may be subtracted as well:
    
    >>> combustion = mixed_reaction - fermentation
    >>> combustion.show()
    Reaction (by mol):
    stoichiometry              reactant    X[%]
    Glucose + O2 -> H2O + CO2  Glucose    20.00
    
    When a Reaction object is multiplied (or divided), a new Reaction object
    with the conversion multiplied (or divided) is returned:
        
    >>> combustion_multiplied = 2 * combustion
    >>> combustion_multiplied.show()
    Reaction (by mol):
    stoichiometry              reactant    X[%]
    Glucose + O2 -> H2O + CO2  Glucose    40.00
    >>> fermentation_divided = fermentation / 2
    >>> fermentation_divided.show()
    Reaction (by mol):
    stoichiometry                  reactant    X[%]
    Glucose + O2 -> Ethanol + CO2  Glucose    35.00
    
    """
    phases = MaterialIndexer.phases
    __slots__ = ('_basis',
                 '_phases',
                 '_chemicals',
                 '_X_index', 
                 '_stoichiometry', 
                 '_X')
    
    def __init__(self, reaction, reactant=None, X=1., 
                 chemicals=None, basis='mol', *,
                 phases=None,
                 check_mass_balance=False,
                 check_atomic_balance=False,
                 correct_atomic_balance=False,
                 correct_mass_balance=False):
        if basis in ('wt', 'mol'):
            self._basis = basis
        else:
            raise ValueError("basis must be either by 'wt' or by 'mol'")
        self.X = X
        chemicals = self._load_chemicals(chemicals)
        if reaction:
            self._phases = phases = phase_tuple(phases) if phases else xprs.get_phases(reaction)
            if phases:
                self._stoichiometry = stoichiometry = xprs.get_stoichiometric_array(reaction, phases, chemicals)
                if reactant is None:
                    _, reactants_index = stoichiometry.negative_index()
                    N_reactants = len(reactants_index)
                    if N_reactants == 1:
                        reactant_index = reactants_index[0]
                    else:
                        raise ValueError('must pass reactant when multiple reactants are involved')
                else:
                    reactant_index = self._chemicals.index(reactant)
                for phase_index, x in enumerate(stoichiometry[:, reactant_index]):
                    if x: break
                self._X_index = (phase_index, reactant_index)
            else:
                self._stoichiometry = prs.get_stoichiometric_array(reaction, chemicals)
                if reactant is None:
                    reactants_index, = self._stoichiometry.negative_index()
                    N_reactants = len(reactants_index)
                    if N_reactants == 1:
                        self._X_index = reactants_index[0]
                    else:
                        raise ValueError('must pass reactant when multiple reactants are involved')
                else:
                    self._X_index = self._chemicals.index(reactant)
            self._rescale()
            if correct_atomic_balance:
                self.correct_atomic_balance()
            else:
                if correct_mass_balance:
                    self.correct_mass_balance()
                elif check_mass_balance:
                    self.check_mass_balance()
                if check_atomic_balance:
                    self.check_atomic_balance()
        else:
            self._phases = ()
            self._stoichiometry = SparseVector.from_size(chemicals.size)
            self._X_index = self._chemicals.index(reactant)
    
    def backwards(self, reactant=None, X=None):
        new = self.copy()
        if reactant is None:
            if new._phases:
                _, reactants_index = new._stoichiometry.positive_index()
                N_reactants = len(reactants_index)
                if N_reactants == 1:
                    new._X_index = reactants_index[0]
                else:
                    raise ValueError('must pass reactant when multiple reactants are involved')
            else:
                reactants_index, = new._stoichiometry.positive_index()
                N_reactants = len(reactants_index)
                if N_reactants == 1:
                    self._X_index = reactants_index[0]
                else:
                    raise ValueError('must pass reactant when multiple reactants are involved')
        else:
            new._X_index = new._chemicals.index(reactant)
        if X is not None: new.X = X
        new._rescale()
        return new
    
    @property
    def reaction_chemicals(self):
        """Return all chemicals involved in the reaction."""
        keys = sorted(self._stoichiometry.nonzero_keys())
        chemicals = self._chemicals
        return [chemicals[i] for i in keys]
    
    def reset_chemicals(self, chemicals):
        if self._chemicals is chemicals: return
        phases = self.phases
        stoichiometry = self._stoichiometry
        reactant = self.reactant
        if phases:
            M, N = stoichiometry.shape
            new_stoichiometry = SparseArray.from_shape([M, chemicals.size])
            IDs = self._chemicals.IDs
            for (i, j), k in stoichiometry.nonzero_items():
                new_stoichiometry[i, chemicals.index(IDs[j])] = k
            phase, reactant = reactant
            X_index = phases.index(phase), chemicals.index(reactant)
        else:
            new_stoichiometry = SparseVector.from_size(chemicals.size)
            IDs = self._chemicals.IDs
            for i, j in stoichiometry.nonzero_items():
                new_stoichiometry[chemicals.index(IDs[i])] = j
            X_index = chemicals.index(reactant)
        self._chemicals = chemicals
        self._stoichiometry = new_stoichiometry
        self._X_index = X_index
    
    def copy(self, basis=None):
        """Return copy of Reaction object."""
        copy = self.__new__(self.__class__)
        copy._basis = self._basis
        copy._phases = self._phases
        copy._stoichiometry = self._stoichiometry.copy()
        copy._X_index = self._X_index
        copy._chemicals = self._chemicals
        copy._X = self._X
        if basis: set_reaction_basis(copy, basis)
        return copy
    
    def has_reaction(self):
        """Return whether any reaction takes place."""
        return bool(self.X and self.stoichiometry.any())
    
    def _math_compatible_reaction(self, rxn, copy=True):
        basis = self.basis
        if copy or basis != rxn._basis: rxn = rxn.copy(basis)
        if self._chemicals is not rxn._chemicals:
            raise ValueError('chemicals must be the same to add/subtract reactions')
        if self._phases != rxn._phases:
            raise ValueError('phases must be the same to add/subtract reactions')
        if self._X_index != rxn._X_index:
            raise ValueError('reactants must be the same to add/subtract reactions')
        return rxn
    
    def __radd__(self, rxn):
        return self + rxn
    
    def __add__(self, rxn):
        if rxn == 0 or rxn is None or not rxn.has_reaction(): return self.copy()
        rxn = self._math_compatible_reaction(rxn)
        stoichiometry = self._stoichiometry * self.X + rxn._stoichiometry * rxn.X
        rxn._stoichiometry = stoichiometry / -(stoichiometry[rxn._X_index])
        rxn.X = self.X + rxn.X
        return rxn
    
    def __iadd__(self, rxn):
        if rxn == 0 or rxn is None or not rxn.has_reaction(): return self
        rxn = self._math_compatible_reaction(rxn, copy=False)
        stoichiometry = self._stoichiometry * self.X + rxn._stoichiometry * rxn.X
        self._stoichiometry = stoichiometry / -(stoichiometry[self._X_index])
        self.X = self.X + rxn.X
        return self
    
    def __mul__(self, num):
        new = self.copy()
        new.X *= float(num)
        return new
    
    def __rmul__(self, num):
        return self.__mul__(num)
    
    def __imul__(self, num):
        self.X *= num
        return self
    
    def __truediv__(self, num):
        return self.__mul__(1./num)  
    
    def __itruediv__(self, num):
        return self.__imul__(1./num) 
    
    def __neg__(self):
        new = self.copy()
        new.X *= -1.
        return new
    
    def __sub__(self, rxn):
        if rxn == 0 or rxn is None or not rxn.has_reaction(): return self
        rxn = self._math_compatible_reaction(rxn)
        stoichiometry = self._stoichiometry*self.X - rxn._stoichiometry*rxn.X
        rxn._stoichiometry = stoichiometry/-(stoichiometry[rxn._X_index])
        rxn.X = self.X - rxn.X
        return rxn
    
    def __isub__(self, rxn):
        if rxn == 0 or rxn is None or not rxn.has_reaction(): return self
        rxn = self._math_compatible_reaction(rxn, copy=False)
        stoichiometry = self._stoichiometry*self.X + rxn._stoichiometry*rxn.X
        self._stoichiometry = stoichiometry/-(stoichiometry[self._X_index])
        self.X = self.X - rxn.X
        return self
    
    def __call__(self, material):
        values, config, original = as_material_array(
            material, self._basis, self._phases, self._chemicals
        )
        self._reaction(values)
        if tmo.reaction.CHECK_FEASIBILITY:
            has_negatives = values.has_negatives()
            if has_negatives:
                negative_index, = values.negative_index()
                negative_values = values[negative_index]
                if negative_values.sum() < -1e-12:
                    X_net = self.X_net()
                    for ID, X in X_net.items():
                        if X > 1.: RuntimeError(f"conversion of '{ID}' is over 100%")
                    if values.ndim == 2:
                        negative_index = values.negative_columns()
                    chemicals = self.chemicals.tuple
                    IDs = [chemicals[i].ID for i in negative_index]
                    if len(IDs) == 1: IDs = repr(IDs[0])
                    raise InfeasibleRegion(f'conversion of {IDs} is over 100%; reaction conversion')
                else:
                    values[negative_index] = 0.
        else:
            fn.remove_negligible_negative_values(values)
        if original is not None: original[:] = values
        if config: material._imol.reset_chemicals(*config)
        
    def force_reaction(self, material):
        """React material ignoring feasibility checks."""
        values, config, original = as_material_array(
            material, self._basis, self._phases, self._chemicals
        )
        self._reaction(values)
        fn.remove_negligible_negative_values(values)
        if original is not None: original[:] = values
        if config: material._imol.reset_chemicals(*config)
    
    def reactant_demand(self, reactant, basis=None, reactant_demand=None):
        """
        Return or set the demand of any reactant (i.e., the reactant's 
        stoichiometric coefficient multiplied by the conversion).

        If this function is used to set the reactant yield,
        the conversion is updated and the stoichiometric coefficient is kept 
        constant.

        Parameters
        ----------
        reactant : str, optional
            ID of the reactant chemical.
        basis : str, optional
            Can be 'mol' or 'wt'. Defaults to the Reaction object's basis.
        reactant_demand: float, optional
            New reactant demand as a number between 0 and 1. If none given, 
            the reactant demand is returned.

        """
        if reactant_demand is None:
            return -self.product_yield(reactant, basis, reactant_demand)
        else:
            self.product_yield(reactant, basis, -reactant_demand)
    
    def product_yield(self, product, basis=None, product_yield=None):
        """
        Return or set the yield of a product per reactant (i.e., the product's
        stoichiometric coefficient multiplied by the conversion).

        If this function is used to set the product yield,
        the conversion is updated and the stoichiometric coefficient is kept 
        constant.

        Parameters
        ----------
        product : str, optional
            ID of the product chemical.
        basis : str, optional
            Can be 'mol' or 'wt'. Defaults to the Reaction object's basis.
        product_yield : float, optional
            New product yield as a number between 0 and 1. If none given, 
            the product yield is returned.
        
        """
        stoichiometry = self._stoichiometry
        if stoichiometry.ndim == 2: 
            stoichiometry = stoichiometry.sum(axis=0)
            reactant_index = self._X_index[1]
        else:
            reactant_index = self._X_index
        product_index = self._chemicals.index(product)
        product_coefficient = stoichiometry[product_index]
        if product_yield is None:
            product_yield = product_coefficient * self.X
            if basis and self.basis != basis:
                chemicals_tuple = self._chemicals.tuple
                MW_reactant = chemicals_tuple[reactant_index].MW
                MW_product = chemicals_tuple[product_index].MW
                if basis == 'wt':
                    product_yield *= MW_product / MW_reactant 
                    assert product_yield <= 1.
                elif basis == 'mol':
                    product_yield *= MW_reactant / MW_product 
                else:
                    raise ValueError("basis must be either 'wt' or 'mol'; "
                                    f"not {repr(basis)}")
            return product_yield
        else:
            X = product_yield / product_coefficient
            if basis and self.basis != basis:
                chemicals_tuple = self._chemicals.tuple
                MW_reactant = chemicals_tuple[reactant_index].MW
                MW_product = chemicals_tuple[product_index].MW
                if basis == 'wt':
                    X *= MW_reactant / MW_product  
                elif basis == 'mol':
                    X *= MW_product / MW_reactant 
                else:
                    raise ValueError("basis must be either 'wt' or 'mol'; "
                                    f"not {repr(basis)}")
            if abs(X) > 1.: 
                raise ValueError("product yield or reactant demand is above the maximum theoretical")
            self.X = X
    
    def adiabatic_reaction(self, stream):
        """
        React stream material adiabatically, accounting for the change in enthalpy
        due to the heat of reaction.
        
        Examples
        --------
        Note how the stream temperature changes after each reaction due to the
        heat of reaction. Adiabatic combustion of hydrogen:
        
        >>> import thermosteam as tmo
        >>> chemicals = tmo.Chemicals(['H2', 'O2', 'H2O'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> reaction = tmo.Reaction('2H2 + O2 -> 2H2O', reactant='H2', X=0.7)
        >>> s1 = tmo.Stream('s1', H2=10, O2=20, H2O=1000, T=373.15, phase='g')
        >>> s2 = tmo.Stream('s2')
        >>> s2.copy_like(s1) # s1 and s2 are the same
        >>> s1.show() # Before reaction
        Stream: s1
        phase: 'g', T: 373.15 K, P: 101325 Pa
        flow (kmol/hr): H2   10
                        O2   20
                        H2O  1e+03
        
        >>> reaction.show()
        Reaction (by mol):
        stoichiometry       reactant    X[%]
        H2 + 0.5 O2 -> H2O  H2         70.00
        
        >>> reaction(s1) 
        >>> s1.show() # After isothermal reaction
        Stream: s1
        phase: 'g', T: 373.15 K, P: 101325 Pa
        flow (kmol/hr): H2   3
                        O2   16.5
                        H2O  1.01e+03
        
        >>> reaction.adiabatic_reaction(s2)
        >>> s2.show() # After adiabatic reaction
        Stream: s2
        phase: 'g', T: 421.6 K, P: 101325 Pa
        flow (kmol/hr): H2   3
                        O2   16.5
                        H2O  1.01e+03
        
        Adiabatic combustion of H2 and CH4:
        
        >>> import thermosteam as tmo
        >>> chemicals = tmo.Chemicals(['H2', 'CH4', 'O2', 'CO2', 'H2O'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> reaction = tmo.ParallelReaction([
        ...    #            Reaction definition          Reactant    Conversion
        ...    tmo.Reaction('2H2 + O2 -> 2H2O',        reactant='H2',  X=0.7),
        ...    tmo.Reaction('CH4 + O2 -> CO2 + 2H2O',  reactant='CH4', X=0.1)
        ... ])
        >>> s1 = tmo.Stream('s1', H2=10, CH4=5, O2=100, H2O=100, T=373.15, phase='g')
        >>> s2 = tmo.Stream('s2')
        >>> s1.show() # Before reaction
        Stream: s1
        phase: 'g', T: 373.15 K, P: 101325 Pa
        flow (kmol/hr): H2   10
                        CH4  5
                        O2   100
                        H2O  100
        
        >>> reaction.show()
        ParallelReaction (by mol):
        index  stoichiometry            reactant    X[%]
        [0]    H2 + 0.5 O2 -> H2O       H2         70.00
        [1]    CH4 + O2 -> CO2 + 2 H2O  CH4        10.00
        
        >>> reaction.adiabatic_reaction(s1)
        >>> s1.show() # After adiabatic reaction
        Stream: s1
        phase: 'g', T: 666.38 K, P: 101325 Pa
        flow (kmol/hr): H2   3
                        CH4  4.5
                        O2   96
                        CO2  0.5
                        H2O  108
        
        Sequential combustion of CH4 and CO:
        
        >>> import thermosteam as tmo
        >>> chemicals = tmo.Chemicals(['CH4', 'CO','O2', 'CO2', 'H2O'], cache=True)
        >>> tmo.settings.set_thermo(chemicals)
        >>> reaction = tmo.SeriesReaction([
        ...     #            Reaction definition                 Reactant       Conversion
        ...     tmo.Reaction('2CH4 + 3O2 -> 2CO + 4H2O',       reactant='CH4',    X=0.7),
        ...     tmo.Reaction('2CO + O2 -> 2CO2',               reactant='CO',     X=0.1)
        ...     ])
        >>> s1 = tmo.Stream('s1', CH4=5, O2=100, H2O=100, T=373.15, phase='g')
        >>> s1.show() # Before reaction
        Stream: s1
        phase: 'g', T: 373.15 K, P: 101325 Pa
        flow (kmol/hr): CH4  5
                        O2   100
                        H2O  100
        
        >>> reaction.show()
        SeriesReaction (by mol):
        index  stoichiometry               reactant    X[%]
        [0]    CH4 + 1.5 O2 -> CO + 2 H2O  CH4        70.00
        [1]    CO + 0.5 O2 -> CO2          CO         10.00
        
        >>> reaction.adiabatic_reaction(s1)
        >>> s1.show() # After adiabatic reaction
        Stream: s1
        phase: 'g', T: 650.01 K, P: 101325 Pa
        flow (kmol/hr): CH4  1.5
                        CO   3.15
                        O2   94.6
                        CO2  0.35
                        H2O  107
        
        """
        if not isinstance(stream, tmo.Stream):
            raise ValueError(f"stream must be a Stream object, not a '{type(stream).__name__}' object")
        Hnet = stream.Hnet
        self(stream)
        stream.H = Hnet - stream.Hf
    
    def _reaction(self, material_array):
        material_array += material_array[self._X_index] * self.X * self._stoichiometry
    
    def X_net(self, indexer=False):
        """Return net reaction conversion of reactants as a dictionary or
        a ChemicalIndexer if indexer is True."""
        X_net = {self.reactant: self.X}
        if indexer:
            chemicals = self.chemicals
            data = chemicals.kwarray(X_net)
            return ChemicalIndexer.from_data(data, NoPhase, chemicals, False)
        else:
            return X_net
    
    @property
    def dH(self):
        """
        Heat of reaction at given conversion. Units are in either
        J/mol-reactant or J/g-reactant; depending on basis.
        
        Warning
        -------
        This property also accounts for:
        * Latent heats (at 298.15 K) when phases are specified.
        * The extent of reaction, `X`.
        * The basis of the reaction (by mol or by wt)

        """
        chemicals = self._chemicals
        stoichiometry = self._stoichiometry.to_array()
        phases = self.phases
        Hfs = chemicals.Hf
        if phases:
            H_latent = np.zeros_like(stoichiometry)
            for i, phase in enumerate(self.phases):
                for j, chemical in enumerate(chemicals):
                    phase_ref = chemical.phase_ref
                    if phase_ref != phase and stoichiometry[i, j] != 0.:
                        if phase_ref == 'l':
                            if phase == 'g':
                                H_latent[i, j] = chemical.Hvap(298.15)
                            elif phase == 's':
                                H_latent[i, j] = -chemical.Hfus
                            else:
                                raise RuntimeError(f"invalid phase '{phase}'")
                        elif phase_ref == 'g':
                            if phase == 'l':
                                H_latent[i, j] = -chemical.Hvap(298.15)
                            elif phase == 's':
                                H_latent[i, j] = -(chemical.Hvap(298.15) + chemical.Hfus)
                            else:
                                raise RuntimeError(f"invalid phase '{phase}'")
                        elif phase_ref == 's':
                            if phase == 'l':
                                H_latent[i, j] = chemical.Hfus
                            elif phase == 'g':
                                H_latent[i, j] = chemical.Hfus + chemical.Hvap(298.15)
                            else:
                                raise RuntimeError(f"invalid phase '{phase}'")
                        else:
                            raise RuntimeError(f"invalid reference phase '{phase_ref}'")
            Hfs = Hfs + H_latent
        if self._basis == 'wt': Hfs = Hfs / self.MWs
        return self.X * (Hfs * stoichiometry).sum()
    
    @property
    def X(self):
        """[float] Reaction conversion as a fraction."""
        return self._X
    @X.setter
    def X(self, X):
        self._X = float(X)
    
    @property
    def stoichiometry(self):
        """[sparse array] Stoichiometry coefficients."""
        return self._stoichiometry
    @property
    def istoichiometry(self):
        """[ChemicalIndexer] Stoichiometry coefficients."""
        stoichiometry = self._stoichiometry
        if stoichiometry.ndim == 1:
            return tmo.indexer.ChemicalIndexer.from_data(self._stoichiometry,
                                                         chemicals=self._chemicals,
                                                         check_data=False)
        else:
            return tmo.indexer.MaterialIndexer.from_data(self._stoichiometry,
                                                         phases=self._phases,
                                                         chemicals=self._chemicals,
                                                         check_data=False)
    
    @property
    def reactant(self):
        """[str] Reactant associated to conversion."""
        if self._phases:
            phase_index, chemical_index = self._X_index
            return self._phases[phase_index], self._chemicals.IDs[chemical_index]
        else:
            return self._chemicals.IDs[self._X_index] 

    @property
    def MWs(self):
        """[1d array] Molecular weights of all chemicals [g/mol]."""
        return self._chemicals.MW

    @property
    def basis(self):
        """{'mol', 'wt'} Basis of reaction"""
        return self._basis
    @basis.setter
    def basis(self, basis):
        set_reaction_basis(self, basis)

    def _get_stoichiometry_by_wt(self):
        """Return stoichiometry by weight."""
        if self._basis == 'mol':
            stoichiometry_by_wt = self._stoichiometry * self.MWs
        else:
            stoichiometry_by_wt = self._stoichiometry
        return stoichiometry_by_wt
        
    def _get_stoichiometry_by_mol(self):
        """Return stoichiometry on a molar basis."""
        if self._basis == 'wt':
            stoichiometry_by_mol = self._stoichiometry / self.MWs
        else:
            stoichiometry_by_mol = self._stoichiometry
        return stoichiometry_by_mol
    
    def mass_balance_error(self):
        """Return error in stoichiometric mass balance. If positive,
        mass is being created. If negative, mass is being destroyed."""
        stoichiometry_by_wt = self._get_stoichiometry_by_wt()
        return stoichiometry_by_wt.sum()
    
    def atomic_balance_error(self):
        """Return a dictionary of errors in stoichiometric atomic balances. 
        If value is positive, the atom is being created. If negative, the atom 
        is being destroyed."""
        stoichiometry_by_mol = self._get_stoichiometry_by_mol()
        formula_array = self.chemicals.formula_array
        unbalanced_array = formula_array @ stoichiometry_by_mol
        return elements.array_to_atoms(unbalanced_array)
    
    def check_mass_balance(self, tol=1e-3):
        """Check that stoichiometric mass balance is correct."""
        error = self.mass_balance_error()
        if abs(error) > tol:
            raise RuntimeError("material stoichiometry is unbalanced by "
                              f"{error} g / mol-reactant")
    
    def check_atomic_balance(self, tol=1e-3):
        """Check that stoichiometric atomic balance is correct."""
        atoms = self.atomic_balance_error()
        if abs(sum(atoms.values())) > tol: 
            raise RuntimeError("atomic stoichiometry is unbalanced by the "
                               "following molar stoichiometric coefficients:\n "
            + "\n ".join([f"{symbol}: {value}" for symbol, value in atoms.items()])
        )
    
    def correct_mass_balance(self, variable=None):
        """
        Make sure mass is not created or destroyed by varying the 
        reactant stoichiometric coefficient.
        """
        if variable:
            index = self.chemicals.get_index(variable)
        else:
            index = self._X_index
        stoichiometry_by_wt = self._get_stoichiometry_by_wt()
        if self.phases: 
            stoichiometry_by_wt = stoichiometry_by_wt.sum(0)
        def f(x):
            stoichiometry_by_wt[index] = x
            return stoichiometry_by_wt.sum()
        
        x = flx.aitken_secant(f, 1)
        if self._basis == 'mol': 
            x /= self.MWs[index]
            if self.phases:
                row = np.where(self._stoichiometry[:, index])
                self._stoichiometry[row, index] = x
            else:
                self._stoichiometry[index] = x 
        self._rescale()
    
    def correct_atomic_balance(self, constants=None):
        """
        Correct stoichiometry coefficients to satisfy atomic balance.
        
        Parameters
        ----------
        constants : str, optional
            IDs of chemicals for which stoichiometric coefficients are held constant.
        
        Examples
        --------
        Balance glucose fermentation to ethanol:
        
        >>> import thermosteam as tmo
        >>> tmo.settings.set_thermo(['Glucose', 'O2', 'CO2', 'Ethanol', 'CH4', 'Water'])
        >>> fermentation = tmo.Reaction('Glucose + O2 -> Ethanol + CO2',
        ...                             reactant='Glucose',  X=0.9)
        >>> fermentation.correct_atomic_balance()
        >>> fermentation.show()
        Reaction (by mol):
        stoichiometry                 reactant    X[%]
        Glucose -> 2 CO2 + 2 Ethanol  Glucose    90.00
        
        Balance methane combustion:
            
        >>> combustion = tmo.Reaction('CH4 + O2 -> Water + CO2',
        ...                           reactant='CH4', X=1)
        >>> combustion.correct_atomic_balance()
        >>> combustion.show()
        Reaction (by mol):
        stoichiometry                reactant    X[%]
        2 O2 + CH4 -> CO2 + 2 Water  CH4       100.00
         
        Balance electrolysis of water (with chemical phases specified):
            
        >>> electrolysis = tmo.Reaction('H2O,l -> H2,g + O2,g',
        ...                             chemicals=tmo.Chemicals(['H2O', 'H2', 'O2']),
        ...                             reactant='H2O', X=1)
        >>> electrolysis.correct_atomic_balance()
        >>> electrolysis.show()
        Reaction (by mol):
        stoichiometry             reactant    X[%]
        H2O,l -> H2,g + 0.5 O2,g  H2O,l     100.00
        
        Note that if the reaction is underspecified, there are infinite
        ways to balance the reaction and a runtime error is raised:
        
        >>> rxn_underspecified = tmo.Reaction('CH4 + Glucose + O2 -> Water + CO2',
        ...                                   reactant='CH4', X=1)
        >>> rxn_underspecified.correct_atomic_balance()
        Traceback (most recent call last):
        RuntimeError: reaction stoichiometry is underspecified; pass the 
        `constants` argument to the `<Reaction>.correct_atomic_balance` method 
        to specify which stoichiometric coefficients to hold constant
        
        Chemical coefficients can be held constant to prevent this error:
        
        >>> rxn_underspecified = tmo.Reaction('CH4 + Glucose + O2 -> Water + CO2',
        ...                                   reactant='CH4', X=1)
        >>> rxn_underspecified.correct_atomic_balance(['Glucose', 'CH4'])
        >>> rxn_underspecified.show()
        Reaction (by mol):
        stoichiometry                            reactant    X[%]
        Glucose + 8 O2 + CH4 -> 7 CO2 + 8 Water  CH4       100.00
        
        """
        stoichiometry_by_mol = self._get_stoichiometry_by_mol()
        stoichiometry_by_mol = stoichiometry_by_mol.to_array()
        phases = self.phases
        if phases:
            stoichiometry_by_mol = stoichiometry_by_mol.sum(0)
        chemicals = self.chemicals
        if constants:
            if isinstance(constants, str): constants = [constants]
            constants = set(constants)
            constant_index = chemicals.indices(constants)
        else:
            constant_index = [self._X_index[1] if phases else self._X_index]
        chemical_index, = np.where(stoichiometry_by_mol)
        chemical_index = np.setdiff1d(chemical_index, constant_index)
        formula_array = chemicals.formula_array
        b = - (formula_array[:, constant_index]
               * stoichiometry_by_mol[constant_index]).sum(1, keepdims=True)
        atomic_bool_index = np.any(formula_array * stoichiometry_by_mol, axis=1)
        atomic_index, = np.where(atomic_bool_index)
        b = b[atomic_index, :]
        A = formula_array[atomic_index, :][:, chemical_index]
        M_atoms, N_chemicals = A.shape
        if M_atoms != N_chemicals:
            x, _, rank, *_ = np.linalg.lstsq(A, b, rcond=None)
            if N_chemicals > rank:
                raise RuntimeError(
                     "reaction stoichiometry is underspecified (i.e. there are "
                     "infinite ways to balance the reaction); pass the "
                     "`constants` argument to the `<Reaction>.correct_atomic_balance` "
                     "method to specify which stoichiometric coefficients to hold constant"
                )
            residual_mass = ((A @ x - b) * self.MWs).sum()
            if residual_mass > 1e-6:
                warn(f'atomic balance was solved with a residual mass error of {residual_mass} g / mol of reactant')
        else:
            x = np.linalg.solve(A, b)
        
        stoichiometry_by_mol[chemical_index] = x.flatten()
        by_wt = self._basis == 'wt'
        stoichiometry = stoichiometry_by_mol * self.MWs if by_wt else stoichiometry_by_mol
        if phases: 
            self._stoichiometry[:] = (self._stoichiometry != 0.) * stoichiometry
        else:
            self._stoichiometry[:] = stoichiometry
        self._rescale()
    
    def _rescale(self):
        """Scale stoichiometry to a per reactant basis."""
        new_scale = -self._stoichiometry[self._X_index]
        if new_scale == 0.:
            raise RuntimeError(f"reactant '{self.reactant}' does not participate in stoichiometric reaction")
        self._stoichiometry /= new_scale
    
    def to_df(self, index=None):
        columns = [f'Stoichiometry (by {self.basis})', 'Reactant', 'Conversion [%]']
        stoichiometry = get_stoichiometric_string(self.stoichiometry, self.phases, self.chemicals)
        reactant = self.reactant
        if self.phases: 
            phase, reactant = reactant
            reactant += ',' + phase
        conversion = 100. * self.X
        df = pd.DataFrame(data=[[stoichiometry, reactant, conversion]], columns=columns, index=[index] if index else None)
        df.index.name = 'Reaction'
        return df
    
    def __repr__(self):
        reaction = get_stoichiometric_string(self.stoichiometry, self.phases, self.chemicals)
        return f"{type(self).__name__}('{reaction}', reactant='{self.reactant}', X={self.X:.3g}, basis={repr(self.basis)})"
    
    def _info(self):
        info = f"{type(self).__name__} (by {self.basis}):"
        rxn = get_stoichiometric_string(self.stoichiometry, self.phases, self.chemicals)
        if self.phases:
            phase, ID = self.reactant
            cmp = ID + ',' + phase
        else:
            cmp = self.reactant
        lrxn = len(rxn)
        lcmp = len(cmp)
        maxrxnlen = max([13, lrxn]) + 2
        maxcmplen = max([8, lcmp]) + 2
        X = self.X
        info += "\nstoichiometry" + " "*(maxrxnlen - 13) + "reactant" + " "*(maxcmplen - 8) + '  X[%]'
        rxn_spaces = " "*(maxrxnlen - lrxn)
        cmp_spaces = " "*(maxcmplen - lcmp)
        info += f"\n{rxn}{rxn_spaces}{cmp}{cmp_spaces}{X*100: >6.2f}"
        return info
    
    def show(self):
        print(self._info())
    _ipython_display_ = show


class ReactionItem(Reaction):
    """
    Create a ReactionItem object from the a ReactionSet and reaction index.
    
    Parameters
    ----------
    rxnset : ReactionSet
    index : int
        Index of reaction.
        
    """
    __slots__ = ('_index', '_parent')
    phases = MaterialIndexer.phases
    
    def __init__(self, rxnset, index):
        self._stoichiometry = rxnset._stoichiometry[index]
        self._phases = rxnset._phases
        self._basis = rxnset._basis
        self._X = rxnset._X
        self._chemicals = rxnset._chemicals
        self._X_index = rxnset._X_index[index]
        self._index = index
        self._parent = rxnset
    
    def reset_chemicals(self, chemicals):
        if self._chemicals is chemicals: return
        parent = self._parent
        parent.reset_chemicals(chemicals)
        index = self._index
        self._stoichiometry = parent._stoichiometry[index]
        self._X_index = parent._X_index[index]
        self._chemicals = chemicals
    
    @property
    def basis(self):
        """{'mol', 'wt'} Basis of reaction"""
        return self._basis
    @basis.setter
    def basis(self, basis):
        raise TypeError('cannot change basis of reaction item')
    
    def copy(self, basis=None):
        """Return copy of Reaction object."""
        copy = Reaction.__new__(Reaction)
        copy._basis = self._basis
        copy._phases = self._phases
        copy._stoichiometry = self._stoichiometry.copy()
        copy._X_index = self._X_index
        copy._chemicals = self._chemicals
        copy._X = self.X
        if basis: set_reaction_basis(copy, basis)
        return copy
    
    @property
    def X(self):
        """[float] Reaction conversion as a fraction."""
        return self._X[self._index]
    @X.setter
    def X(self, X):
        self._X[self._index] = X
        

class ReactionSet:
    """
    Create a ReactionSet that contains all reactions and conversions as an array.
    
    Parameters
    ----------
    reactions : Iterable[Reaction]
    
    """
    __slots__ = (*Reaction.__slots__, '_parent_index')
    copy = Reaction.copy
    phases = MaterialIndexer.phases
    _get_stoichiometry_by_mol = Reaction._get_stoichiometry_by_mol
    _get_stoichiometry_by_wt = Reaction._get_stoichiometry_by_wt
    force_reaction = Reaction.force_reaction
    adiabatic_reaction = Reaction.adiabatic_reaction
    __call__ = Reaction.__call__
    
    def __init__(self, reactions):
        if not reactions: raise ValueError('no reactions passed')
        phases_set = set([i.phases for i in reactions])
        if len(phases_set) > 1:
            raise ValueError('all reactions must implement the same phases')
        self._phases, = phases_set
        chemicals = {i.chemicals for i in reactions}
        try: self._chemicals, = chemicals
        except: raise ValueError('all reactions must have the same chemicals')
        basis = {i.basis for i in reactions}
        try: self._basis, = basis
        except: raise ValueError('all reactions must have the same basis')
        self._stoichiometry = [i._stoichiometry for i in reactions]
        self._X = np.array([i.X for i in reactions])
        X_index = [i._X_index for i in reactions]
        self._X_index = tuple(X_index) if self._phases else np.array(X_index)
        
    def __getitem__(self, index):
        stoichiometry = self._stoichiometry[index]
        if stoichiometry.__class__ is list:
            rxnset = self.__new__(self.__class__)
            rxnset._basis = self._basis
            rxnset._phases = self._phases
            rxnset._stoichiometry = stoichiometry
            rxnset._X = self._X[index]
            rxnset._X_index = self._X_index[index]
            rxnset._chemicals = self._chemicals
            rxnset._parent_index = (self, index)
            return rxnset
        else:
            return ReactionItem(self, index)
    
    def _rescale(self):
        """Scale stoichiometry to a per reactant basis."""
        stoichiometry = self._stoichiometry
        for i, index in enumerate(self._X_index): 
            row = stoichiometry[i]
            row /= -row[index]
    
    def reset_chemicals(self, chemicals):
        if chemicals is self._chemicals: return
        if hasattr(self, '_parent_index'):
            parent, index = self._parent_index
            parent.reset_chemicals(chemicals)
            self._stoichiometry = parent._stoichiometry[index]
            self._X_index = parent._X_index[index]
            self._chemicals = parent._chemicals
            return
        phases = self.phases
        stoichiometry = self._stoichiometry
        reactants = self.reactants
        if phases:
            new_stoichiometry = []
            for stoic in stoichiometry:
                A, B = stoic.shape
                new_stoic = SparseArray.from_shape([A, chemicals.size])
                new_stoichiometry.append(new_stoic)
                IDs = self._chemicals.IDs
                for i in range(A):
                    for j in range(B):
                        value = stoic[i, j]
                        if value: new_stoic[i, chemicals.index(IDs[j])] = value
            X_index = [(phases.index(i), chemicals.index(j)) for i, j in reactants]
        else:
            A, B = np.array(stoichiometry).shape
            new_stoichiometry = SparseArray.from_shape([A, chemicals.size])
            IDs = self._chemicals.IDs
            for i in range(A):
                for j in range(B):
                    value = stoichiometry[i, j]
                    if value: new_stoichiometry[i, chemicals.index(IDs[j])] = value
            X_index = [chemicals.index(i) for i in reactants]
        self._chemicals = chemicals
        self._stoichiometry = new_stoichiometry
        self._X_index = X_index
    
    @property
    def reaction_chemicals(self):
        """Return all chemicals involved in the reaction."""
        if self._phases:
            keys = []
            for i in self._stoichiometry: keys.extend(i.nonzero_keys())
            keys = sorted(set(keys))
        else:
            keys = sorted(self._stoichiometry.nonzero_keys())
        chemicals = self._chemicals
        return [chemicals[i] for i in keys]
    
    @property
    def basis(self):
        """{'mol', 'wt'} Basis of reaction"""
        return self._basis
    @basis.setter
    def basis(self, basis):
        raise TypeError('cannot change basis of reaction set')
    
    @property
    def X(self):
        """[1d array] Reaction conversions."""
        return self._X
    @X.setter
    def X(self, X):
        """[1d array] Reaction conversions."""
        if X is not self._X: self._X[:] = X
    
    @property
    def chemicals(self):
        """[Chemicals] Chemicals corresponing to each entry in the stoichiometry array."""
        return self._chemicals
    @property
    def stoichiometry(self):
        """[2d array] Stoichiometry coefficients."""
        return self._stoichiometry
    
    @property
    def reactants(self):
        """tuple[str] Reactants associated to conversion."""
        IDs = self._chemicals.IDs
        phases = self._phases
        X_index = self._X_index
        if phases:
            return tuple([(phases[i], IDs[j]) for i,j in X_index])
        else:
            return tuple([IDs[i] for i in X_index])
    
    @property
    def MWs(self):
        """[2d array] Molecular weights of all chemicals."""
        return self._chemicals.MW[np.newaxis, :]
    
    def to_df(self, index=None):
        columns = [f'Stoichiometry (by {self.basis})', 'Reactant', 'Conversion [%]']
        chemicals = self._chemicals
        phases = self._phases
        rxns = [get_stoichiometric_string(i, phases, chemicals) for i in self._stoichiometry]
        cmps = [ID + ',' + phase for phase, ID in self.reactants] if phases else self.reactants
        Xs = 100. * self.X
        data = list([i for i in zip(rxns, cmps, Xs) if i[2]])
        df = pd.DataFrame(data, columns=columns, index=index if index else None)
        if isinstance(self, ParallelReaction):
            df.index.name = 'Parallel reaction'
        elif isinstance(self, SeriesReaction):
            df.index.name = 'Reaction in series'
        return df
    
    def __repr__(self):
        return f"{type(self).__name__}([{', '.join([repr(i).replace('ReactionItem', 'Reaction') for i in self])}])"
    
    def _info(self, index_name='index'):
        info = f"{type(self).__name__} (by {self.basis}):"
        chemicals = self._chemicals
        phases = self._phases
        length = len
        string = str
        rxns = [get_stoichiometric_string(i, phases, chemicals) for i in self._stoichiometry]
        maxrxnlen = max([13, *[length(i) for i in rxns]]) + 2
        cmps = [ID + ',' + phase for phase, ID in self.reactants] if phases else self.reactants
        maxcmplen = max([8, *[length(i) for i in cmps]]) + 2
        Xs = self.X
        N = len(Xs)
        maxnumspace = max(length(string(N)) + 1, len(index_name))
        info += f"\n{index_name}" + " "*(max(2, length(string(N)))) + "stoichiometry" + " "*(maxrxnlen - 13) + "reactant" + " "*(maxcmplen - 8) + '  X[%]'
        for N, rxn, cmp, X in zip(range(N), rxns, cmps, Xs):
            rxn_spaces = " "*(maxrxnlen - length(rxn))
            cmp_spaces = " "*(maxcmplen - length(cmp))
            num = string(N)
            numspace = (maxnumspace - length(num)) * " "
            info += f"\n[{N}]{numspace}{rxn}{rxn_spaces}{cmp}{cmp_spaces}{X*100: >6.2f}"
        return info
    _ipython_display_ = show = Reaction.show
        

class ParallelReaction(ReactionSet):
    """
    Create a ParallelReaction object from Reaction objects. When called, 
    it returns the change in material due to all parallel reactions.
    
    Parameters
    ----------
    reactions : Iterable[Reaction]
    
    Examples
    --------
    Run two reactions in parallel:
    
    >>> import thermosteam as tmo
    >>> chemicals = tmo.Chemicals(['H2', 'Ethanol', 'CH4', 'O2', 'CO2', 'H2O'], cache=True)
    >>> tmo.settings.set_thermo(chemicals)
    >>> kwargs = dict(phases='lg', correct_atomic_balance=True)
    >>> reaction = tmo.ParallelReaction([
    ...    #            Reaction definition                    Reactant             Conversion
    ...    tmo.Reaction('H2,g + O2,g -> 2H2O,g',               reactant='H2',       X=0.7, **kwargs),
    ...    tmo.Reaction('Ethanol,l + O2,g -> CO2,g + 2H2O,g',  reactant='Ethanol',  X=0.1, **kwargs)
    ... ])
    >>> reaction.reactants # Note that reactants are tuples of phase and ID pairs.
    (('g', 'H2'), ('l', 'Ethanol'))
    
    >>> reaction.show()
    ParallelReaction (by mol):
    index  stoichiometry                            reactant     X[%]
    [0]    H2,g + 0.5 O2,g -> H2O,g                 H2,g        70.00
    [1]    3 O2,g + Ethanol,l -> 2 CO2,g + 3 H2O,g  Ethanol,l   10.00
    
    >>> s1 = tmo.MultiStream('s1', T=373.15, 
    ...                      l=[('Ethanol', 10)],
    ...                      g=[('H2', 10), ('CH4', 5), ('O2', 100), ('H2O', 10)])
    >>> s1.show() # Before reaction
    MultiStream: s1
    phases: ('g', 'l'), T: 373.15 K, P: 101325 Pa
    flow (kmol/hr): (g) H2       10
                        CH4      5
                        O2       100
                        H2O      10
                    (l) Ethanol  10
    
    >>> reaction(s1)
    >>> s1.show() # After isothermal reaction
    MultiStream: s1
    phases: ('g', 'l'), T: 373.15 K, P: 101325 Pa
    flow (kmol/hr): (g) H2       3
                        CH4      5
                        O2       93.5
                        CO2      2
                        H2O      20
                    (l) Ethanol  9
    
    Reaction items are accessible:
    
    >>> reaction[0].show()
    ReactionItem (by mol):
    stoichiometry             reactant    X[%]
    H2,g + 0.5 O2,g -> H2O,g  H2,g       70.00
    
    Note that changing the conversion of a reaction item changes the 
    conversion of its parent reaction set:
        
    >>> reaction[0].X = 0.5
    >>> reaction.show()
    ParallelReaction (by mol):
    index  stoichiometry                            reactant     X[%]
    [0]    H2,g + 0.5 O2,g -> H2O,g                 H2,g        50.00
    [1]    3 O2,g + Ethanol,l -> 2 CO2,g + 3 H2O,g  Ethanol,l   10.00
    
    Reactions subsets can be made as well:
        
    >>> reaction[:1].show()
    ParallelReaction (by mol):
    index  stoichiometry             reactant    X[%]
    [0]    H2,g + 0.5 O2,g -> H2O,g  H2,g       50.00
    
    Get net reaction conversion of reactants as a material indexer:
        
    >>> mi = reaction.X_net(indexer=True)
    >>> mi.show()
    MaterialIndexer:
     (g) H2        0.5
     (l) Ethanol   0.1
    >>> mi['g', 'H2']
    0.5
    
    If no phases are specified for a reaction set, the `X_net` property returns
    a ChemicalIndexer:
    
    >>> kwargs = dict(correct_atomic_balance=True)
    >>> reaction = tmo.ParallelReaction([
    ...    #            Reaction definition            Reactant             Conversion
    ...    tmo.Reaction('H2 + O2 -> 2H2O',             reactant='H2',       X=0.7, **kwargs),
    ...    tmo.Reaction('Ethanol + O2 -> CO2 + 2H2O',  reactant='Ethanol',  X=0.1, **kwargs)
    ... ])
    >>> ci = reaction.X_net(indexer=True)
    >>> ci.show()
    ChemicalIndexer:
     H2       0.7
     Ethanol  0.1
    >>> ci['H2']
    0.7
    
    """
    __slots__ = ()
    
    def _reaction(self, material_array):
        reacted = self._X * np.array([material_array[i] for i in self._X_index], float)
        for X, stoichiometry in zip(reacted, self._stoichiometry):
            material_array += X * stoichiometry

    def reduce(self):
        """
        Return a new Parallel reaction object that combines reaction 
        with the same reactant together, reducing the number of reactions.
        """
        rxn_dict = {i: [] for i in set(self._X_index)}
        for i in self: rxn_dict[i._X_index].append(i)
        for key, rxns in rxn_dict.items():
            rxn, *rxns = rxns
            rxn = rxn.copy()
            for i in rxns: rxn += i
            rxn_dict[key] = rxn 
        return self.__class__(rxn_dict.values())
            
    def X_net(self, indexer=False):
        """Return net reaction conversion of reactants as a dictionary or
        a ChemicalIndexer if indexer is True."""
        X_net = {}
        for i, j in zip(self.reactants, self.X):
            if i in X_net:
                X_net[i] += j
            else:
                X_net[i] = j
        if indexer:
            chemicals = self.chemicals
            phases = self.phases
            if phases:
                phases = [i[0] for i in X_net]
                mi = MaterialIndexer(phases=phases, chemicals=chemicals)
                for i,j in X_net.items(): mi[i] = j
                return mi                
            else:
                data = chemicals.kwarray(X_net)
                return ChemicalIndexer.from_data(data, NoPhase, chemicals, False)
        else:
            return X_net


class SeriesReaction(ReactionSet):
    """
    Create a ParallelReaction object from Reaction objects. When called, 
    it returns the change in material due to all reactions in series.
    
    Parameters
    ----------
    reactions : Iterable[Reaction]
    
    
    """
    __slots__ = ()

    def reduce(self):
        raise TypeError('cannot reduce a SeriesReation object, only '
                        'ParallelReaction objects are reducible')

    def _reaction(self, material_array):
        for i, j, k in zip(self._X_index, self.X, self._stoichiometry):
            material_array += material_array[i] * j * k

    def X_net(self, indexer=False):
        """Return net reaction conversion of reactants as a dictionary or
        a ChemicalIndexer if indexer is True."""
        X_net = {}
        for i, j in zip(self.reactants, self.X):
            if i in X_net:
                X_net[i] += (1 - X_net[i]) * j
            else:
                X_net[i] = j
        if indexer:
            chemicals = self.chemicals
            data = chemicals.kwarray(X_net)
            return ChemicalIndexer.from_data(data, NoPhase, chemicals, False)
        else:
            return X_net

class ReactionSystem:
    """
    Create a ReactionSystem object that can react a stream across a series of
    reactions.
    
    Parameters
    ----------
    *reactions : Reaction, ParallelReaction, or SeriesReaction
        All reactions within the reaction system.
    
    Examples
    --------
    Create a reaction system for cellulosic fermentation of biomass:
    
    >>> from thermosteam import Rxn, RxnSys, PRxn, SRxn, settings, Chemical, Stream
    >>> cal2joule = 4.184
    >>> Glucan = Chemical('Glucan', search_db=False, formula='C6H10O5', Hf=-233200*cal2joule, phase='s', default=True)
    >>> Glucose = Chemical('Glucose', phase='s')
    >>> CO2 = Chemical('CO2', phase='g')
    >>> HMF = Chemical('HMF', search_ID='Hydroxymethylfurfural', phase='l', default=True)
    >>> Biomass = Glucose.copy(ID='Biomass')
    >>> settings.set_thermo(['Water', 'Ethanol', 'LacticAcid', HMF, Glucose, Glucan, CO2, Biomass])
    
    >>> saccharification = PRxn([
    ...     Rxn('Glucan + H2O -> Glucose', reactant='Glucan', X=0.9),
    ...     Rxn('Glucan -> HMF + 2H2O', reactant='Glucan', X=0.025)
    ... ])
    >>> fermentation = SRxn([
    ...     Rxn('Glucose -> 2LacticAcid', reactant='Glucose', X=0.03),
    ...     Rxn('Glucose -> 2Ethanol + 2CO2', reactant='Glucose', X=0.95),
    ... ])
    >>> cell_growth = Rxn('Glucose -> Biomass', reactant='Glucose', X=1.0)
    >>> cellulosic_rxnsys = RxnSys(saccharification, fermentation, cell_growth)
    >>> cellulosic_rxnsys.show()
    ReactionSystem:
    [0]  ParallelReaction (by mol):
         index  stoichiometry              reactant    X[%]
         [0]    Water + Glucan -> Glucose  Glucan     90.00
         [1]    Glucan -> 2 Water + HMF    Glucan      2.50
    [1]  SeriesReaction (by mol):
         index  stoichiometry                 reactant    X[%]
         [0]    Glucose -> 2 LacticAcid       Glucose     3.00
         [1]    Glucose -> 2 Ethanol + 2 CO2  Glucose    95.00
    [2]  Reaction (by mol):
         stoichiometry       reactant    X[%]
         Glucose -> Biomass  Glucose   100.00
    
    Compute the flux of glucan through saccharification reactions:
    
    >>> feed = Stream('feed', Glucan=1.0, Water=5.0)
    >>> cellulosic_rxnsys.reactant_flux(feed, index=0)
    0.925
    
    Compute the flux of glucan through glucan to glucose saccharification:
    
    >>> cellulosic_rxnsys.reactant_flux(feed, index=0, subindex=0)
    0.9
    
    Compute the flux of glucose through cell growth:
        
    >>> cellulosic_rxnsys.reactant_flux(feed, index=2)
    0.04365
    
    Compute the flux of glucose through ethanol production:
    
    >>> cellulosic_rxnsys.reactant_flux(feed, index=1, subindex=1)
    0.8293
    
    Notice how reacting the stream leads to ethanol production equivalent
    of 2x the glucose flux to that reaction:
    
    >>> cellulosic_rxnsys(feed)
    >>> feed.show()
    Stream: feed
    phase: 'l', T: 298.15 K, P: 101325 Pa
    flow (kmol/hr): Water       4.15
                    Ethanol     1.66
                    LacticAcid  0.054
                    HMF         0.025
                    Glucan      0.075
                    CO2         1.66
                    Biomass     0.0437
    
    """
    __slots__ = ('_reactions',
                 '_basis',
                 '_chemicals',
                 '_phases')
    
    def __init__(self, *reactions, basis=None):
        if not reactions: raise ValueError('Reactions cannot be empty')
        self._reactions = reactions
        try: self._phases, = set([i._phases for i in reactions])
        except: raise ValueError('all reactions must have the same phases')
        try: self._chemicals, = set([i._chemicals for i in reactions])
        except: raise ValueError('all reactions must have the same chemicals')
        try: self._basis, = set([i._basis for i in reactions])
        except: raise ValueError('all reactions must have the same basis')
        
    force_reaction = Reaction.force_reaction
    adiabatic_reaction = Reaction.adiabatic_reaction
    chemicals = Reaction.chemicals
    __call__ = Reaction.__call__
    show = Reaction.show
    
    def __iter__(self):
        return iter(self.reactions)
    
    def __getitem__(self, index):
        return self._reactions[index]
    
    @property
    def X(self):
        return [i.X for i in self._reactions]
    @X.setter
    def X(self, X):
        for i, j in zip(self._reactions, X): i.X = j
    
    @property
    def reactions(self):
        return self._reactions
    
    def _reaction(self, material):
        basis = self._basis
        for i in self._reactions: 
            if i._basis != basis: raise RuntimeError('not all reactions have the same basis')
            i._reaction(material)
        
    def X_net(self, indexer=False):
        """Return net reaction conversion of reactants as a dictionary or
        a ChemicalIndexer if indexer is True."""
        X_net = {}
        for rxn in self._reactions:
            for i, j in rxn.X_net().items():
                if i in X_net:
                    X_net[i] += (1 - X_net[i]) * j
                else:
                    X_net[i] = j
        if indexer:
            chemicals = self._chemicals
            data = chemicals.kwarray(X_net)
            return ChemicalIndexer.from_data(data, NoPhase, chemicals, False)
        else:
            return X_net
        
    def reactant_flux(self, material, index, subindex=None):
        """
        Return the amount of reactant being reacted in a reaction.
        
        Parameters
        ----------
        material : Stream or array
            Material entering reaction system.
        index : int
            Index of reaction to calculate reactant flux.
        subindex : int, optional
            Subindex of reaction to calculate reactant flux.    
        
        """
        values, config, original = as_material_array(
            material, self._basis, self._phases, self._chemicals
        )
        preconverted_material = values if original else values.copy()
        reactions = self.reactions
        for i, rxn in enumerate(reactions):
            if i == index: break
            rxn(preconverted_material)
        reaction = reactions[index]
        if subindex is not None:
            if isinstance(reaction, SeriesReaction):
                reactions = reaction
                for i, rxn in enumerate(reactions):
                    if i == subindex: break
                    rxn(preconverted_material)
            reaction = reaction[subindex]
            X_index = reaction._X_index
            return reaction.X * preconverted_material[X_index]
        elif isinstance(reaction, SeriesReaction):
            raise ValueError('must pass subindex if the index refers to a SeriesReaction object')
        elif isinstance(reaction, ParallelReaction):
            X_index = reaction._X_index
            return (reaction.X * preconverted_material[X_index]).sum()
        else:
            X_index = reaction._X_index
            return reaction.X * preconverted_material[X_index]
        
        
    def __repr__(self):
        return f"{type(self).__name__}({', '.join([repr(i) for i in self.reactions])})"
    
    def _info(self):
        indexed_info = f"{type(self).__name__}:"
        isa = isinstance
        infos = [(i._info() if isa(i, ReactionSet) else i._info()) 
                 for i in self._reactions]
        N = len(infos)
        maxnumspace = len(str(N)) + 2
        info_dim = '\n' + " " * (maxnumspace + 2)
        for index, info in enumerate(infos):
            num = str(index) 
            numspace = (maxnumspace - len(num)) * " "
            info = info.replace('\n', info_dim)
            indexed_info += f"\n[{num}]{numspace}{info}"
        return indexed_info    
        
    _ipython_display_ = show = Reaction.show

# Short-hand conventions
Rxn = Reaction
RxnI = ReactionItem
RxnS = ReactionSet
PRxn = ParallelReaction
SRxn = SeriesReaction
RxnSys = ReactionSystem



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from scipy.optimize import shgo
import flexsolve as flx
from numba import njit
from warnings import warn
from ..exceptions import InfeasibleRegion, NoEquilibrium
from . import binary_phase_fraction as binary
from .equilibrium import Equilibrium
from .dew_point import DewPointCache
from .bubble_point import BubblePointCache
from .fugacity_coefficients import IdealFugacityCoefficients
from .poyinting_correction_factors import MockPoyintingCorrectionFactors
from . import activity_coefficients as ac
from .. import functional as fn
from ..utils import Cache
import numpy as np

__all__ = ('VLE', 'VLECache')

# @njit(cache=True)
def xV_iter(xV, pcf_Psat_over_P_phi, T, P, z, z_light, z_heavy, f_gamma, gamma_args):
    xV = xV.copy()
    x = xV[:-1]
    V = xV[-1]
    x[x < 0.] = 0.
    x = fn.normalize(x)
    Ks = pcf_Psat_over_P_phi * f_gamma(x, T, *gamma_args)
    V = binary.solve_phase_fraction_Rashford_Rice(z, Ks, V, z_light, z_heavy)
    if V < 0.: V = 0.
    elif V > 1.: V = 1.
    xV[-1] = V
    xV[:-1] = z/(1. + V * (Ks - 1.))
    return xV

@njit(cache=True)    
def xV_iter_2n(xV, pcf_Psat_over_P_phi, T, P, z, f_gamma, gamma_args):
    xV = xV.copy()
    x = xV[:-1]
    V = xV[-1]
    x[x < 0.] = 0.
    x = fn.normalize(x)
    Ks = pcf_Psat_over_P_phi * f_gamma(x, T, *gamma_args)
    V = binary.compute_phase_fraction_2N(z, Ks)
    if V < 0.: V = 0.
    elif V > 1.: V = 1.
    xV[-1] = V
    xV[:-1] = z/(1. + V * (Ks - 1.))
    return xV

@njit(cache=True)
def xV_iter_3n(xV, pcf_Psat_over_P_phi, T, P, z, f_gamma, gamma_args):
    xV = xV.copy()
    x = xV[:-1]
    V = xV[-1]
    x[x < 0.] = 0.
    x = fn.normalize(x)
    Ks = pcf_Psat_over_P_phi * f_gamma(x, T, *gamma_args)
    V = binary.compute_phase_fraction_3N(z, Ks)
    if V < 0.: V = 0.
    elif V > 1.: V = 1.
    xV[-1] = V
    xV[:-1] = z/(1. + V * (Ks - 1.))
    return xV

def set_flows(vapor_mol, liquid_mol, index, vapor_data, total_data):
    vapor_mol[index] = vapor_data
    liquid_mol[index] = total_data - vapor_data

class VLE(Equilibrium, phases='lg'):
    """
    Create a VLE object that performs vapor-liquid equilibrium when called.
        
    Parameters
    ----------
    imol=None : :class:`~thermosteam.indexer.MaterialIndexer`, optional
        Molar chemical phase data is stored here.
    thermal_condition=None : :class:`~thermosteam.ThermalCondition`, optional
        Temperature and pressure results are stored here.
    thermo=None : :class:`~thermosteam.Thermo`, optional
        Themodynamic property package for equilibrium calculations.
        Defaults to `thermosteam.settings.get_thermo()`.
    bubble_point_cache=None : :class:`~thermosteam.utils.Cache`, optional
        Cache to retrieve bubble point object.
    dew_point_cache=None : :class:`~thermosteam.utils.Cache`, optional
        Cache to retrieve dew point object
    
    Examples
    --------
    First create a VLE object:
    
    >>> from thermosteam import indexer, equilibrium, settings
    >>> settings.set_thermo(['Water', 'Ethanol', 'Methanol', 'Propanol'], cache=True)
    >>> imol = indexer.MolarFlowIndexer(
    ...             l=[('Water', 304), ('Ethanol', 30)],
    ...             g=[('Methanol', 40), ('Propanol', 1)])
    >>> vle = equilibrium.VLE(imol)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Methanol', 40), ('Propanol', 1)],
            l=[('Water', 304), ('Ethanol', 30)]),
        thermal_condition=ThermalCondition(T=298.15, P=101325))
    
    Equilibrium given vapor fraction and pressure:
    
    >>> vle(V=0.5, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 126.7), ('Ethanol', 26.38), ('Methanol', 33.49), ('Propanol', 0.8958)],
            l=[('Water', 177.3), ('Ethanol', 3.622), ('Methanol', 6.509), ('Propanol', 0.1042)]),
        thermal_condition=ThermalCondition(T=363.85, P=101325))
    
    Equilibrium given temperature and pressure:
    
    >>> vle(T=363.88, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 127.4), ('Ethanol', 26.41), ('Methanol', 33.54), ('Propanol', 0.8968)],
            l=[('Water', 176.6), ('Ethanol', 3.59), ('Methanol', 6.456), ('Propanol', 0.1032)]),
        thermal_condition=ThermalCondition(T=363.88, P=101325))
    
    Equilibrium given enthalpy and pressure:
    
    >>> H = vle.thermo.mixture.xH(vle.imol, T=363.88, P=101325)
    >>> vle(H=H, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 127.4), ('Ethanol', 26.41), ('Methanol', 33.54), ('Propanol', 0.8968)],
            l=[('Water', 176.6), ('Ethanol', 3.59), ('Methanol', 6.456), ('Propanol', 0.1032)]),
        thermal_condition=ThermalCondition(T=363.88, P=101325))
    
    Equilibrium given entropy and pressure:
    
    >>> S = vle.thermo.mixture.xS(vle.imol, T=363.88, P=101325)
    >>> vle(S=S, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 127.4), ('Ethanol', 26.41), ('Methanol', 33.54), ('Propanol', 0.8968)],
            l=[('Water', 176.6), ('Ethanol', 3.59), ('Methanol', 6.456), ('Propanol', 0.1032)]),
        thermal_condition=ThermalCondition(T=363.88, P=101325))
    
    Equilibrium given vapor fraction and temperature:
    
    >>> vle(V=0.5, T=363.88)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 126.7), ('Ethanol', 26.38), ('Methanol', 33.49), ('Propanol', 0.8958)],
            l=[('Water', 177.3), ('Ethanol', 3.622), ('Methanol', 6.509), ('Propanol', 0.1042)]),
        thermal_condition=ThermalCondition(T=363.88, P=101431))
    
    Equilibrium given enthalpy and temperature:
    
    >>> vle(H=H, T=363.88)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 127.4), ('Ethanol', 26.41), ('Methanol', 33.54), ('Propanol', 0.8968)],
            l=[('Water', 176.6), ('Ethanol', 3.59), ('Methanol', 6.456), ('Propanol', 0.1032)]),
        thermal_condition=ThermalCondition(T=363.88, P=101325))
    
    Non-partitioning heavy and gaseous chemicals also affect VLE. Calculation 
    are repeated with non-partitioning chemicals:
        
    >>> from thermosteam import indexer, equilibrium, settings, Chemical
    >>> O2 = Chemical('O2', phase='g')
    >>> Glucose = Chemical('Glucose', phase='l', default=True)
    >>> settings.set_thermo(['Water', 'Ethanol', 'Methanol', 'Propanol', O2, Glucose], cache=True)
    >>> imol = indexer.MolarFlowIndexer(
    ...             l=[('Water', 304), ('Ethanol', 30), ('Glucose', 5)],
    ...             g=[('Methanol', 40), ('Propanol', 1), ('O2', 10)])
    >>> vle = equilibrium.VLE(imol)
    >>> vle(T=363.88, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 159.5), ('Ethanol', 27.65), ('Methanol', 35.63), ('Propanol', 0.9337), ('O2', 10)],
            l=[('Water', 144.5), ('Ethanol', 2.352), ('Methanol', 4.369), ('Propanol', 0.0663), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=363.88, P=101325))
    
    >>> vle(V=0.5, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 126.7), ('Ethanol', 26.38), ('Methanol', 33.52), ('Propanol', 0.8957), ('O2', 10)],
            l=[('Water', 177.3), ('Ethanol', 3.618), ('Methanol', 6.478), ('Propanol', 0.1043), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=362.47, P=101325))
    
    >>> H = vle.thermo.mixture.xH(vle.imol, T=363.88, P=101325)
    >>> vle(H=H, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 127.4), ('Ethanol', 26.42), ('Methanol', 33.58), ('Propanol', 0.8968), ('O2', 10)],
            l=[('Water', 176.6), ('Ethanol', 3.583), ('Methanol', 6.421), ('Propanol', 0.1032), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=362.51, P=101325))
    
    >>> S = vle.thermo.mixture.xS(vle.imol, T=363.88, P=101325)
    >>> vle(S=S, P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 128.2), ('Ethanol', 26.45), ('Methanol', 33.63), ('Propanol', 0.8979), ('O2', 10)],
            l=[('Water', 175.8), ('Ethanol', 3.548), ('Methanol', 6.365), ('Propanol', 0.1021), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=362.54, P=101325))
    
    >>> vle(V=0.5, T=363.88)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 126.7), ('Ethanol', 26.38), ('Methanol', 33.49), ('Propanol', 0.8958), ('O2', 10)],
            l=[('Water', 177.3), ('Ethanol', 3.622), ('Methanol', 6.509), ('Propanol', 0.1042), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=363.88, P=106841))
    
    >>> vle(H=H, T=363.88)
    >>> vle 
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 126.7), ('Ethanol', 26.38), ('Methanol', 33.49), ('Propanol', 0.8958), ('O2', 10)],
            l=[('Water', 177.3), ('Ethanol', 3.622), ('Methanol', 6.51), ('Propanol', 0.1042), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=363.88, P=106842))

    >>> vle(S=S, T=363.88)
    >>> vle 
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 128.1), ('Ethanol', 26.45), ('Methanol', 33.6), ('Propanol', 0.8978), ('O2', 10)],
            l=[('Water', 175.9), ('Ethanol', 3.555), ('Methanol', 6.399), ('Propanol', 0.1022), ('Glucose', 5)]),
        thermal_condition=ThermalCondition(T=363.88, P=106562))

    The presence of a non-partitioning gaseous chemical will result in some
    evaporation, even if the tempeture is below the saturated bubble point:
        
    >>> from thermosteam import indexer, equilibrium, settings, Chemical
    >>> O2 = Chemical('O2', phase='g')
    >>> settings.set_thermo(['Water', O2], cache=True)
    >>> imol = indexer.MolarFlowIndexer(
    ...             l=[('Water', 30)],
    ...             g=[('O2', 10)])
    >>> vle = equilibrium.VLE(imol)
    >>> vle(T=300., P=101325)
    >>> vle
    VLE(imol=MolarFlowIndexer(
            g=[('Water', 0.3617), ('O2', 10)],
            l=[('Water', 29.64)]),
        thermal_condition=ThermalCondition(T=300.00, P=101325))

    """
    __slots__ = (
        'method', # [str] Method for solving equilibrium.
        '_T', # [float] Temperature [K].
        '_P', # [float] Pressure [Pa].
        '_H_hat', # [float] Specific enthalpy [kJ/kg].
        '_S_hat', # [float] Specific entropy [kJ/K/kg].
        '_V', # [float] Molar vapor fraction.
        '_dew_point', # [DewPoint] Solves for dew point.
        '_bubble_point', # [BubblePoint] Solves for bubble point.
        '_x', # [1d array] Vapor composition.
        '_y', # [1d array] Liquid composition.
        '_z_last', # tuple[1d array] Last bulk composition.
        '_phi', # [FugacityCoefficients] Estimates fugacity coefficients of gas.
        '_pcf', # [PoyintingCorrectionFactors] Estimates the PCF of a liquid.
        '_gamma', # [ActivityCoefficients] Estimates activity coefficients of a liquid.
        '_liquid_mol', # [1d array] Liquid molar data.
        '_vapor_mol', # [1d array] Vapor molar data.
        '_phase_data', # tuple[str, 1d array] Phase-data pairs.
        '_v',  # [1d array] Molar vapor data in equilibrium.
        '_index', # [1d array] Index of chemicals in equilibrium.
        '_F_mass', # [float] Total mass data.
        '_chemical', # [Chemical] Single chemical in equilibrium.
        '_mol_vle', # [1d array] Moles of chemicals in VLE calculations.
        '_N', # [int] Number of chemicals in equilibrium.
        '_z', # [1d array] Molar composition of chemicals in equilibrium
        '_z_norm', # [1d array] Normalized molar composition of chemicals in equilibrium
        '_z_light', # [1d array] Molar composition of light chemicals not included in equilibrium calculation.
        '_z_heavy', # [1d array] Molar composition of heavy chemicals not included in equilibrium calculation.
        '_nonzero', # [1d array(bool)] Chemicals present in the mixture
        '_F_mol', # [float] Total moles of chemicals (accounting for dissociation).
        '_F_mol_vle', # [float] Total moles in equilibrium.
        '_F_mol_light', # [float] Total moles of gas chemicals not included in equilibrium calculation.
        '_F_mol_heavy', # [float] Total moles of heavy chemicals not included in equilibrium calculation.
        '_dew_point_cache', # [Cache] Retrieves the DewPoint object if arguments are the same.
        '_bubble_point_cache' # [Cache] Retrieves the BubblePoint object if arguments are the same.
    )
    maxiter = 20
    T_tol = 5e-8
    P_tol = 1.
    H_hat_tol = 1e-6
    S_hat_tol = 1e-6
    V_tol = 1e-6
    x_tol = 1e-9
    y_tol = 1e-9
    default_method = 'fixed-point'
    
    def __init__(self, imol=None, thermal_condition=None,
                 thermo=None, bubble_point_cache=None, dew_point_cache=None):
        self.method = self.default_method
        self._T = self._P = self._H_hat = self._V = 0
        self._dew_point_cache = dew_point_cache or DewPointCache()
        self._bubble_point_cache = bubble_point_cache or BubblePointCache()
        super().__init__(imol, thermal_condition, thermo)
        self._x = None
        self._z_last = None
        self._nonzero = None
        self._index = ()
    
    def __call__(self, *, T=None, P=None, V=None, H=None, S=None, x=None, y=None):
        """
        Perform vapor-liquid equilibrium.

        Parameters
        ----------
        T : float, optional
            Operating temperature [K].
        P : float, optional
            Operating pressure [Pa].
        V : float, optional
            Molar vapor fraction.
        H : float, optional
            Enthalpy [kJ/hr].
        S : float, optional
            Entropy [kJ/hr/K]
        x : float, optional
            Molar composition of liquid (for binary mixtures).
        y : float, optional
            Molar composition of vapor (for binary mixtures).
        
        Notes
        -----
        You may only specify two of the following parameters: P, H, T, V, x, and y.
        Additionally, If x or y is specified, the other parameter must be either
        P or T (e.g., x and V is invalid).
        
        """
        ### Decide what kind of equilibrium to run ###
        T_spec = T is not None
        P_spec = P is not None
        V_spec = V is not None
        H_spec = H is not None
        S_spec = S is not None
        x_spec = x is not None
        y_spec = y is not None
        N_specs = (T_spec + P_spec + V_spec + H_spec + S_spec + x_spec + y_spec)
        assert N_specs == 2, ("must pass two and only two of the following "
                              "specifications: T, P, V, H, S, x, y")
        
        # Run equilibrium
        if T_spec:
            if P_spec:
                try:
                    self.set_thermal_condition(T, P)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.T = T
                    thermal_condition.P = P
            elif V_spec:
                try:
                    self.set_TV(T, V)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.T = T
            elif H_spec:
                self.set_TH(T, H)
            elif S_spec:
                self.set_TS(T, S)
            elif x_spec:
                self.set_Tx(T, np.asarray(x))
            else: # y_spec
                self.set_Ty(T, np.asarray(y))
        elif P_spec:
            if V_spec:
                try:
                    self.set_PV(P, V)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.P = P
            elif H_spec:
                try:
                    self.set_PH(P, H, stacklevel=1)
                except NoEquilibrium:
                    thermal_condition = self._thermal_condition
                    thermal_condition.P = P
            elif S_spec:
                try:
                    self.set_PS(P, S, stacklevel=1)
                except:
                    try:
                        self.set_PS(P, S, stacklevel=1)
                    except NoEquilibrium:
                        thermal_condition = self._thermal_condition
                        thermal_condition.P = P
            elif x_spec:
                self.set_Px(P, np.asarray(x))
            else: # y_spec
                self.set_Py(P, np.asarray(y))
        elif S_spec: # pragma: no cover
            if y_spec:
                raise NotImplementedError('specification S and y is invalid')
            elif x_spec:
                raise NotImplementedError('specification S and x is invalid')
            elif H_spec:
                raise NotImplementedError('specification H and S is invalid')
            else: # V_spec
                raise NotImplementedError('specification V and S not implemented')
        elif H_spec: # pragma: no cover
            if y_spec:
                raise NotImplementedError('specification H and y is invalid')
            elif x_spec:
                raise NotImplementedError('specification H and x is invalid')
            else: # V_spec
                raise NotImplementedError('specification V and H not implemented')
        elif V_spec: # pragma: no cover
            if y_spec:
                raise ValueError("specification V and y is invalid")
            else: # x_spec
                raise ValueError("specification V and x is invalid")
        else: # pragma: no cover
            raise ValueError("can only pass either 'x' or 'y' arguments, not both")
    
    def _setup(self):
        # Get flow rates
        imol = self._imol
        self._phase_data = tuple(imol)
        self._liquid_mol = liquid_mol = imol['l']
        self._vapor_mol = vapor_mol = imol['g']
        mol = liquid_mol + vapor_mol
        nonzero = mol.nonzero_keys()
        chemicals = self.chemicals
        if self._nonzero == nonzero:
            index = self._index
            reset = False
        else:
            # Set up indices for both equilibrium and non-equilibrium species
            index = chemicals.get_vle_indices(nonzero)
            eq_chems = chemicals.tuple
            eq_chems = [eq_chems[i] for i in index]
            reset = True     
            self._nonzero = set(nonzero)
            self._index = index
        
        # Get overall composition
        if not mol.any(): raise NoEquilibrium('no chemicals to perform equilibrium')
        self._F_mass = (chemicals.MW * mol).sum()
        self._mol_vle = mol_vle = mol[index]

        # Set light and heavy keys
        LNK_index = chemicals._light_indices
        HNK_index = chemicals._heavy_indices
        vapor_mol[HNK_index] = 0
        vapor_mol[LNK_index] = light_mol = mol[LNK_index]
        liquid_mol[LNK_index] = 0
        liquid_mol[HNK_index] = heavy_mol = mol[HNK_index]
        self._F_mol_light = F_mol_light = light_mol.sum()
        self._F_mol_heavy = F_mol_heavy = (heavy_mol * chemicals._heavy_solutes).sum()
        self._F_mol_vle = F_mol_vle = mol_vle.sum()
        self._F_mol = F_mol = F_mol_vle + F_mol_light + F_mol_heavy
        if F_mol == 0.: raise NoEquilibrium('no chemicals to perform equilibrium')
        self._z = mol_vle / F_mol
        self._z_light = z_light = F_mol_light / F_mol
        self._z_heavy = z_heavy = F_mol_heavy / F_mol
        self._z_norm = mol_vle / F_mol_vle
        N = len(index)
        if N:
            N += z_light > 0.
            N += z_heavy > 0.
        self._N = N
        if reset:
            if N == 0:
                self._phi = self._gamma = self._pcf = self._dew_point = self._bubble_point = None
            elif N == 1:
                self._chemical, = eq_chems
            else:
                # Set equilibrium objects
                thermo = self._thermo
                self._bubble_point = bp = self._bubble_point_cache(eq_chems, thermo)
                self._dew_point = self._dew_point_cache(eq_chems, thermo)
                self._pcf = bp.pcf
                self._gamma = bp.gamma
                self._phi = bp.phi

    @property
    def imol(self):
        return self._imol
    @property
    def thermal_condition(self):
        return self._thermal_condition
    
    ### Single component equilibrium case ###
        
    def _set_thermal_condition_chemical(self, T, P):
        chemical = self._chemical
        if T >= chemical.Tc: 
            self._liquid_mol[self._index] = 0
            self._vapor_mol[self._index] = self._mol_vle
        else:
            # Either liquid or gas
            Psat = chemical.Psat(T)
            tol = 1e-3
            if P < Psat - tol:
                self._liquid_mol[self._index] = 0
                self._vapor_mol[self._index] = self._mol_vle
            elif P > Psat + tol:
                self._liquid_mol[self._index] = self._mol_vle
                self._vapor_mol[self._index] = 0
    
    def _set_TV_chemical(self, T, V):
        # Set vapor fraction
        self._T = self._thermal_condition.T = self._chemical.Psat(T)
        self._vapor_mol[self._index] = V * self._mol_vle
        self._liquid_mol[self._index] = self._mol_vle - self._vapor_mol[self._index]
        
    def _set_PV_chemical(self, P, V):
        # Set vapor fraction
        self._T = self._thermal_condition.T = self._chemical.Tsat(P, check_validity=False)
        self._vapor_mol[self._index] = self._mol_vle * V
        self._liquid_mol[self._index] = self._mol_vle - self._vapor_mol[self._index]
        
    def _set_PH_chemical(self, P, H): 
        mol = self._mol_vle
        index = self._index
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        thermo = self._thermo
        phase_data = self._phase_data
        mixture = thermo.mixture
        chemical = self._chemical
        
        # Set temperature in equilibrium
        self._T = self._thermal_condition.T = T = chemical.Tsat(P, check_validity=False) 
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = mixture.xH(phase_data, T, P)
        if H >= H_dew:
            self._thermal_condition.T = mixture.xsolve_T_at_HP(phase_data, H, T, P)
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = mixture.xH(phase_data, T, P)
        if H <= H_bubble:
            self._thermal_condition.T = mixture.xsolve_T_at_HP(phase_data, H, T, P)
            return
        
        # Adjust vapor fraction accordingly
        V = (H - H_bubble)/(H_dew - H_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _set_TH_chemical(self, T, H):
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        mixture = self._thermo.mixture
        phase_data = self._phase_data
        
        # Set Pressure in equilibrium
        self._thermal_condition.P = P = self._chemical.Psat(T)
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = mixture.xH(phase_data, T, P)
        if H >= H_dew:
            raise NotImplementedError('cannot solve for pressure yet')
            self._thermal_condition.P = mixture.xsolve_P_at_HT(phase_data, H, T, P)
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = mixture.xH(phase_data, T, P)
        if H <= H_bubble:
            raise NotImplementedError('cannot solve for pressure yet')
            self._thermal_condition.P = mixture.xsolve_P_at_HT(phase_data, H, T, P)
            return
        
        # Adjust vapor fraction accordingly
        V = (H - H_bubble)/(H_dew - H_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _set_PS_chemical(self, P, S): 
        mol = self._mol_vle
        index = self._index
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        thermo = self._thermo
        phase_data = self._phase_data
        mixture = thermo.mixture
        chemical = self._chemical
        
        # Set temperature in equilibrium
        self._T = self._thermal_condition.T = T = chemical.Tsat(P, check_validity=False)
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        S_dew = mixture.xS(phase_data, T, P)
        if S >= S_dew:
            self._thermal_condition.T = mixture.xsolve_T_at_SP(phase_data, S, T, P)
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        S_bubble = mixture.xS(phase_data, T, P)
        if S <= S_bubble:
            self._thermal_condition.T = mixture.xsolve_T_at_SP(phase_data, S, T, P)
            return
        
        # Adjust vapor fraction accordingly
        V = (S - S_bubble)/(S_dew - S_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _set_TS_chemical(self, T, S):
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        mixture = self._thermo.mixture
        phase_data = self._phase_data
        
        # Set Pressure in equilibrium
        self._thermal_condition.P = P = self._chemical.Psat(T)
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        S_dew = mixture.xS(phase_data, T, P)
        if S >= S_dew:
            raise NotImplementedError('cannot solve for pressure yet')
            self._thermal_condition.P = mixture.xsolve_P_at_ST(phase_data, S, T, P)
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        S_bubble = mixture.xS(phase_data, T, P)
        if S <= S_bubble:
            raise NotImplementedError('cannot solve for pressure yet')
            self._thermal_condition.P = mixture.xsolve_P_at_ST(phase_data, S, T, P)
            return
        
        # Adjust vapor fraction accordingly
        V = (S - S_bubble)/(S_dew - S_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _lever_rule(self, x, y):
        split_frac = (self._z[0]-x[0])/(y[0]-x[0])
        if not -0.00001 < split_frac < 1.00001:
            raise InfeasibleRegion('phase composition')
        if split_frac > 1:
            split_frac = 1
        elif split_frac < 0:
            split_frac = 0
        self._vapor_mol[self._index] = v = self._F_mol * split_frac * y
        self._liquid_mol[self._index] = self._mol_vle - v
    
    def set_Tx(self, T, x):
        self._setup()
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify x'
        self._thermal_condition.P, y = self._bubble_point.solve_Py(x, T)
        self._lever_rule(x, y)
    
    def set_Px(self, P, x):
        self._setup()
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify x'
        self._thermal_condition.T, y = self._bubble_point.solve_Ty(x, P) 
        self._lever_rule(x, y)
        
    def set_Ty(self, T, y):
        self._setup()
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify y'
        self._thermal_condition.P, x = self._dew_point.solve_Px(y, T)
        self._lever_rule(x, y)
    
    def set_Py(self, P, y):
        self._setup()
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify y'
        self._thermal_condition.T, x = self._dew_point.solve_Tx(y, P) 
        self._lever_rule(x, y)
        
    def set_thermal_condition(self, T, P):
        self._setup()
        thermal_condition = self._thermal_condition
        self._T = thermal_condition.T = T
        self._P = thermal_condition.P = P
        if self._N == 0: return
        if self._N == 1: return self._set_thermal_condition_chemical(T, P)
        # Check if there is equilibrium
        P_dew, x_dew = self._dew_point.solve_Px(self._z, T)
        if P <= P_dew and not self._F_mol_heavy:
            self._vapor_mol[self._index] = self._mol_vle
            self._liquid_mol[self._index] = 0
            return
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._z, T)
        if P >= P_bubble and not self._F_mol_light:
            self._vapor_mol[self._index] = 0
            self._liquid_mol[self._index] = self._mol_vle
            return
        # Guess composition in the vapor is a
        # weighted average of bubble/dew points
        dP = (P_bubble - P_dew)
        V = (P - P_dew) / dP if dP > 1. else 0.5
        self._refresh_v(V, y_bubble)
        set_flows(self._vapor_mol, self._liquid_mol, self._index, self._solve_v(T, P), self._mol_vle)
        
    def set_TV(self, T, V):
        self._setup()
        mol = self._mol_vle
        thermal_condition = self._thermal_condition
        thermal_condition.T = self._T = T
        if self._N == 0: raise RuntimeError('no chemicals present to perform VLE')
        if self._N == 1: return self._set_TV_chemical(T, V)
        if self._F_mol_heavy and V == 1.: V = 1. - 1e-3
        if self._F_mol_light and V == 0.: V = 1e-3
        if V == 1:
            P_dew, x_dew = self._dew_point.solve_Px(self._z, T)
            self._vapor_mol[self._index] = self._mol_vle
            self._liquid_mol[self._index] = 0
            thermal_condition.P = P_dew
        elif V == 0:
            P_bubble, y_bubble = self._bubble_point.solve_Py(self._z, T)
            self._vapor_mol[self._index] = 0
            self._liquid_mol[self._index] = self._mol_vle
            thermal_condition.P = P_bubble
        else:
            P_bubble, y_bubble = self._bubble_point.solve_Py(self._z, T)
            P_dew, x_dew = self._dew_point.solve_Px(self._z, T)
            self._refresh_v(V, y_bubble)
            if self._F_mol_light: P_bubble = self._bubble_point.Pmax
            if self._F_mol_heavy: P_dew = self._bubble_point.Pmin
            
            V_bubble = self._V_err_at_P(P_bubble, 0.)
            if V_bubble > V:
                F_mol_vapor = self._F_mol * V
                v = y_bubble * F_mol_vapor
                mask = v > mol
                v[mask] = mol[mask]
                P = P_bubble
            else:
                V_dew = self._V_err_at_P(P_dew, 0.)
                if V_dew < V:
                    l = x_dew * self._F_mol * (1. - V)
                    mask = l > mol 
                    l[mask] = mol[mask]
                    v = mol - l
                    P = P_dew
                else:
                    P = flx.IQ_interpolation(
                        self._V_err_at_P,
                        P_bubble, P_dew, V_bubble - V, V_dew - V,
                        self._P, self.P_tol, self.V_tol,
                        (V,), checkiter=False, checkbounds=False,
                        maxiter=self.maxiter,
                    )
                    v = self._v
            
            self._P = thermal_condition.P = P
            set_flows(self._vapor_mol, self._liquid_mol, self._index, v, mol)
            self._H_hat = self.mixture.xH(self._phase_data, T, P) / self._F_mass

    def set_TH(self, T, H):
        self._setup()
        if self._N == 0: raise RuntimeError('no chemicals present to perform VLE')
        if self._N == 1: return self._set_TH_chemical(T, H)
        self._T = T
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        phase_data = self._phase_data
        
        # Check if super heated vapor
        P_dew, x_dew = self._dew_point.solve_Px(self._z, T)
        if self._F_mol_heavy: P_dew = 0.5 * P_dew + 0.5 * self._bubble_point.Pmin
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = self.mixture.xH(phase_data, T, P_dew)
        dH_dew = (H - H_dew)
        if dH_dew >= 0:
            raise NotImplementedError('cannot solve for pressure yet')

        # Check if subcooled liquid
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._z, T)
        if self._F_mol_light: P_bubble = 2 * P_bubble
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = self.mixture.xH(phase_data, T, P_bubble)
        dH_bubble = (H - H_bubble)
        if dH_bubble <= 0:
            raise NotImplementedError('cannot solve for pressure yet')

        # Guess overall vapor fraction, and vapor flow rates
        V = dH_bubble/(H_dew - H_bubble)
        
        # Guess composition in the vapor is a weighted average of boiling points
        self._refresh_v(V, y_bubble)
        F_mass = self._F_mass
        H_hat = H/F_mass
        P = flx.IQ_interpolation(
            self._H_hat_err_at_P,
            P_bubble, P_dew,
            H_bubble/F_mass - H_hat, H_dew/F_mass - H_hat,
            self._P, self.P_tol, self.H_hat_tol,
            (H_hat,), checkiter=False, checkbounds=False,
            maxiter=self.maxiter,
        )
        self._P = self._thermal_condition.P = P   
        self._thermal_condition.T = T
    
    def set_TS(self, T, S):
        self._setup()
        if self._N == 0: raise RuntimeError('no chemicals present to perform VLE')
        if self._N == 1: return self._set_TS_chemical(T, S)
        self._T = T
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        phase_data = self._phase_data
        
        # Check if super heated vapor
        P_dew, x_dew = self._dew_point.solve_Px(self._z, T)
        if self._F_mol_heavy: P_dew = 0.5 * P_dew + 0.5 * self._bubble_point.Pmin
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        S_dew = self.mixture.xS(phase_data, T, P_dew)
        dS_dew = (S - S_dew)
        if dS_dew >= 0:
            raise NotImplementedError('cannot solve for pressure yet')

        # Check if subcooled liquid
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._z, T)
        if self._F_mol_light: P_bubble = 2 * P_bubble
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        S_bubble = self.mixture.xS(phase_data, T, P_bubble)
        dS_bubble = (S - S_bubble)
        if dS_bubble <= 0:
            raise NotImplementedError('cannot solve for pressure yet')

        # Guess overall vapor fraction, and vapor flow rates
        V = dS_bubble/(S_dew - S_bubble)
        
        # Guess composition in the vapor is a weighted average of boiling points
        self._refresh_v(V, y_bubble)
        F_mass = self._F_mass
        S_hat = S/F_mass
        P = flx.IQ_interpolation(
            self._S_hat_err_at_P,
            P_bubble, P_dew,
            S_bubble/F_mass - S_hat, S_dew/F_mass - S_hat,
            self._P, self.P_tol, self.S_hat_tol,
            (S_hat,), checkiter=False, checkbounds=False,
            maxiter=self.maxiter,
        )
        self._P = self._thermal_condition.P = P   
        self._thermal_condition.T = T
    
    def set_PV(self, P, V):
        self._setup()
        self._thermal_condition.P = self._P = P
        if self._N == 0: raise RuntimeError('no chemicals present to perform VLE')
        if self._N == 1: return self._set_PV_chemical(P, V)
        
        # Setup bounderies
        thermal_condition = self._thermal_condition
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        if self._F_mol_heavy and V == 1.: V = 1. - 1e-3
        if self._F_mol_light and V == 0.: V = 1e-3
        if V == 1:
            T_dew, x_dew = self._dew_point.solve_Tx(self._z, P)
            vapor_mol[index] = mol
            liquid_mol[index] = 0
            thermal_condition.T = T_dew
        elif V == 0 and not self._F_mol_light:
            T_bubble, y_bubble = self._bubble_point.solve_Ty(self._z, P)
            vapor_mol[index] = 0
            liquid_mol[index] = mol
            thermal_condition.T = T_bubble
        else:
            T_dew, x_dew = self._dew_point.solve_Tx(self._z, P)
            T_bubble, y_bubble = self._bubble_point.solve_Ty(self._z, P)
            self._refresh_v(V, y_bubble)
            if self._F_mol_heavy: T_dew = 0.9 * T_dew + 0.1 * self._dew_point.Tmax
            if self._F_mol_light: T_bubble = 0.9 * T_bubble + 0.1 * self._bubble_point.Tmin
            V_bubble = self._V_err_at_T(T_bubble, 0.)
            if V_bubble > V:
                F_mol_vapor = self._F_mol * V
                v = y_bubble * F_mol_vapor
                mask = v > mol
                v[mask] = mol[mask]
                T = T_bubble
            else:
                V_dew = self._V_err_at_T(T_dew, 0.)
                if V_dew < V:
                    l = x_dew * self._F_mol * (1. - V)
                    mask = l > mol 
                    l[mask] = mol[mask]
                    v = mol - l
                    T = T_dew
                else:
                    T = flx.IQ_interpolation(
                        self._V_err_at_T,
                        T_bubble, T_dew, V_bubble - V, V_dew - V,
                        self._T, self.T_tol, self.V_tol,
                        (V,), checkiter=False, checkbounds=False,
                        maxiter=self.maxiter,
                    )
                
                    v = self._v
            self._T = thermal_condition.T = T
            set_flows(vapor_mol, liquid_mol, index, v, mol)
            self._H_hat = self.mixture.xH(self._phase_data, T, P)/self._F_mass
    
    def set_PS(self, P, S, stacklevel=0):
        self._setup()
        thermal_condition = self._thermal_condition
        thermal_condition.P = self._P = P
        if self._N == 0: 
            thermal_condition.T = self.mixture.xsolve_T_at_SP(
                self._phase_data, S, thermal_condition.T, P
            )
            return
        if self._N == 1: return self._set_PS_chemical(P, S)
        
        # Setup bounderies
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        
        # Check if subcooled liquid
        T_bubble, y_bubble = self._bubble_point.solve_Ty(self._z, P)
        if self._F_mol_light: T_bubble = 0.9 * T_bubble + 0.1 * self._bubble_point.Tmin
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        S_bubble = self.mixture.xS(self._phase_data, T_bubble, P)
        dS_bubble = S - S_bubble
        if dS_bubble <= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_SP(self._phase_data, S, T_bubble, P)
            return
        
        # Check if super heated vapor
        T_dew, x_dew = self._dew_point.solve_Tx(self._z, P)
        if T_dew <= T_bubble: 
            T_dew, T_bubble = T_bubble, T_dew
            T_dew += 0.5
            T_bubble -= 0.5
        if self._F_mol_heavy: T_dew = 0.9 * T_dew + 0.1 * self._dew_point.Tmax
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        S_dew = self.mixture.xS(self._phase_data, T_dew, P)
        dS_dew = S - S_dew
        if dS_dew >= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_SP(self._phase_data, S, T_dew, P)
            return
        
        # Guess T, overall vapor fraction, and vapor flow rates
        V = dS_bubble/(S_dew - S_bubble)
        self._refresh_v(V, y_bubble)
        
        F_mass = self._F_mass
        S_hat = S/F_mass
        S_hat_bubble = self._S_hat_err_at_T(T_bubble, 0.)
        if (S_hat_bubble:=self._S_hat_err_at_T(T_bubble, 0.)) > S_hat:
            T = T_bubble
        elif (S_hat_dew:=self._S_hat_err_at_T(T_dew, 0.)) < S_hat:
            T = T_dew
        else:
            T = flx.IQ_interpolation(
                self._S_hat_err_at_T, T_bubble, T_dew, 
                S_hat_bubble - S_hat, S_hat_dew - S_hat,
                self._T, self.T_tol, self.S_hat_tol,
                (S_hat,), checkiter=False, checkbounds=False,
                maxiter=self.maxiter
            )
        # Make sure energy balance is correct by vaporizing a fraction
        # of the liquid or condensing a fraction of the vapor
        self._T = thermal_condition.T = T
        mol_liq = liquid_mol.copy()
        mol_liq[:] = 0.
        mol_liq[index] = liquid_mol[index]
        mol_gas= vapor_mol.copy()
        mol_gas[:] = 0.
        mol_gas[index] = vapor_mol[index]
        S_gas = self.mixture.S('g', mol_gas, T, P)
        S_liq = self.mixture.S('l', mol_liq, T, P)
        S_current = self.mixture.xS(self._phase_data, T, P)
        if S_current > S: # Condense a fraction of the vapor
            # Energy balance: S = f * S_condense + S_current
            S_condense = self.mixture.S('l', mol_gas, T, P) - S_gas
            try:
                f = (S - S_current) / S_condense
            except: # Floating point errors
                f = 0
            else:
                if f < 0.:
                    f = 0.
                elif f > 0.:
                    if f > 1.: f = 1.
                    condensed = f * mol_gas[index]
                    liquid_mol[index] += condensed
                    vapor_mol[index] -= condensed 
                else:
                    f = 0.
        else: # Vaporize a fraction of the liquid
            # Energy balance: S = f * S_vaporise + S_current
            S_vaporise = self.mixture.S('g', mol_liq, T, P) - S_liq
            try:
                f = (S - S_current) / S_vaporise
            except: # Floating point errors
                f = 0
            else:
                if f < 0.:
                    f = 0.
                elif f > 0.:
                    if f > 1.: f = 1.
                    vaporised = f * mol_liq[index]
                    vapor_mol[index] += vaporised
                    liquid_mol[index] -= vaporised
                else:
                    f = 0.
        if f == 0. or f == 1.:
            self._T = thermal_condition.T = self.mixture.xsolve_T_at_SP(
                self._phase_data, S, T, P
            )
        self._S_hat = S_hat
    
    def set_PH(self, P, H, stacklevel=0):
        self._setup()
        thermal_condition = self._thermal_condition
        thermal_condition.P = self._P = P
        if self._N == 0: 
            thermal_condition.T = self.mixture.xsolve_T_at_HP(
                self._phase_data, H, thermal_condition.T, P
            )
            return
        if self._N == 1: return self._set_PH_chemical(P, H)
        
        # Setup bounderies
        index = self._index
        mol = self._mol_vle
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        
        # Check if subcooled liquid
        T_bubble, y_bubble = self._bubble_point.solve_Ty(self._z, P)
        if self._F_mol_light: T_bubble = 0.9 * T_bubble + 0.1 * self._bubble_point.Tmin
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = self.mixture.xH(self._phase_data, T_bubble, P)
        dH_bubble = H - H_bubble
        if dH_bubble <= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_HP(self._phase_data, H, T_bubble, P)
            return
        
        # Check if super heated vapor
        T_dew, x_dew = self._dew_point.solve_Tx(self._z, P)
        if T_dew <= T_bubble: 
            T_dew, T_bubble = T_bubble, T_dew
            T_dew += 0.5
            T_bubble -= 0.5
        if self._F_mol_heavy: T_dew = 0.9 * T_dew + 0.1 * self._dew_point.Tmax
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = self.mixture.xH(self._phase_data, T_dew, P)
        dH_dew = H - H_dew
        if dH_dew >= 0:
            thermal_condition.T = self.mixture.xsolve_T_at_HP(self._phase_data, H, T_dew, P)
            return
        
        # Guess T, overall vapor fraction, and vapor flow rates
        V = dH_bubble/(H_dew - H_bubble)
        self._refresh_v(V, y_bubble)
        
        F_mass = self._F_mass
        H_hat = H/F_mass
        
        if (H_hat_bubble:=self._H_hat_err_at_T(T_bubble, 0.)) > H_hat:
            T = T_bubble
        elif (H_hat_dew:=self._H_hat_err_at_T(T_dew, 0.)) < H_hat:
            T = T_dew
        else:
            T = flx.IQ_interpolation(
                self._H_hat_err_at_T, T_bubble, T_dew, 
                H_hat_bubble - H_hat, H_hat_dew - H_hat,
                self._T, self.T_tol, self.H_hat_tol,
                (H_hat,), checkiter=False, checkbounds=False,
                maxiter=self.maxiter
            )
                
        # Make sure energy balance is correct by vaporizing a fraction
        # of the liquid or condensing a fraction of the vapor
        self._T = thermal_condition.T = T
        mol_liq = liquid_mol.copy()
        mol_liq[:] = 0.
        mol_liq[index] = liquid_mol[index]
        mol_gas= vapor_mol.copy()
        mol_gas[:] = 0.
        mol_gas[index] = vapor_mol[index]
        H_gas = self.mixture.H('g', mol_gas, T, P)
        H_liq = self.mixture.H('l', mol_liq, T, P)
        H_current = self.mixture.xH(self._phase_data, T, P)
        if H_current > H: # Condense a fraction of the vapor
            # Energy balance: H = f * H_condense + H_current
            H_condense = self.mixture.H('l', mol_gas, T, P) - H_gas
            try:
                f = (H - H_current) / H_condense
            except: # Floating point errors
                f = 0
            else:
                if f < 0.:
                    f = 0.
                elif f > 0.:
                    if f > 1.: f = 1.
                    condensed = f * mol_gas[index]
                    liquid_mol[index] += condensed
                    vapor_mol[index] -= condensed 
                else:
                    f = 0
        else: # Vaporize a fraction of the liquid
            # Energy balance: H = f * H_vaporise + H_current
            H_vaporise = self.mixture.H('g', mol_liq, T, P) - H_liq
            try:
                f = (H - H_current) / H_vaporise
            except: # Floating point errors
                f = 0
            else:
                if f < 0.:
                    f = 0.
                elif f > 0.:
                    if f > 1.: f = 1.
                    vaporised = f * mol_liq[index]
                    vapor_mol[index] += vaporised
                    liquid_mol[index] -= vaporised
                else:
                    f = 0
        if f == 0. or f == 1.:
            self._T = thermal_condition.T = self.mixture.xsolve_T_at_HP(
                self._phase_data, H, T, P
            )
        self._H_hat = H_hat
    
    def _estimate_v(self, V, y_bubble):
        return (V*self._z_norm + (1-V)*y_bubble) * V * self._F_mol_vle
    
    def _refresh_v(self, V, y_bubble):
        self._v = v = self._estimate_v(V, y_bubble)
        self._V = V
        self._y = fn.normalize(v, v.sum() + self._F_mol_light)
        z_last = self._z_last
        try:
            reload_cache = self._x is None or np.abs(z_last - self._z).sum() > 0.001
        except:
            reload_cache = True
        if reload_cache:
            l = self._mol_vle - v
            self._x = fn.normalize(l, l.sum() + self._F_mol_heavy)
    
    def _H_hat_err_at_T(self, T, H_hat):
        set_flows(self._vapor_mol, self._liquid_mol, self._index, 
                  self._solve_v(T, self._P), self._mol_vle)
        self._H_hat = self.mixture.xH(self._phase_data, T, self._P)/self._F_mass
        return self._H_hat - H_hat
    
    def _H_hat_err_at_P(self, P, H_hat):
        set_flows(self._vapor_mol, self._liquid_mol, self._index,
                  self._solve_v(self._T, P), self._mol_vle)
        self._H_hat = self.mixture.xH(self._phase_data, self._T, P)/self._F_mass
        return self._H_hat - H_hat
    
    def _S_hat_err_at_T(self, T, S_hat):
        set_flows(self._vapor_mol, self._liquid_mol, self._index,
                  self._solve_v(T, self._P), self._mol_vle)
        self._S_hat = self.mixture.xS(self._phase_data, T, self._P)/self._F_mass
        return self._S_hat - S_hat
    
    def _S_hat_err_at_P(self, P, S_hat):
        set_flows(self._vapor_mol, self._liquid_mol, self._index,
                  self._solve_v(self._T, P), self._mol_vle)
        self._S_hat = self.mixture.xS(self._phase_data, self._T, P)/self._F_mass
        return self._S_hat - S_hat
    
    def _V_err_at_P(self, P, V):
        return self._solve_v(self._T , P).sum()/self._F_mol_vle - V
    
    def _V_err_at_T(self, T, V):
        return self._solve_v(T, self._P).sum()/self._F_mol_vle  - V
    
    def _y_iter(self, y, pcf_Psat_over_P, T, P):
        phi = self._phi(y, T, P)
        gamma = self._gamma
        x = self._x
        pcf_Psat_over_P_phi = pcf_Psat_over_P / phi
        N = self._N
        z = self._z
        if N > 3 or self._z_light or self._z_heavy:
            f = xV_iter
            args = (pcf_Psat_over_P_phi, T, P, z, self._z_light, 
                    self._z_heavy, gamma.f, gamma.args)
        elif N == 2:
            f = xV_iter_2n
            args = (pcf_Psat_over_P_phi, T, P, z, gamma.f, gamma.args)
        elif N == 3:
            f = xV_iter_3n
            args = (pcf_Psat_over_P_phi, T, P, z, gamma.f, gamma.args)
        xV = np.zeros(x.size + 1)
        xV[:-1] = x
        xV[-1] = self._V
        xV = flx.aitken(f, xV, self.x_tol, args, checkiter=False, 
                        checkconvergence=False, convergenceiter=5,
                        maxiter=self.maxiter)
        x = xV[:-1]
        self._V = V = xV[-1]
        x[x < 1e-32] = 1e-32
        self._x = xV[:-1] = x = fn.normalize(x)
        if V == 0:
            Ks = 0
        else:
            Ks = (z / x - 1) / V + 1.
        self._z_last = z
        v = self._F_mol * V * x * Ks    
        return fn.normalize(v, v.sum() + self._F_mol_light)
    
    def _solve_v(self, T, P):
        """Solve for vapor mol"""
        method = self.method
        if method == 'shgo':
            gamma = self._gamma
            phi = self._phi
            Psats = np.array([i(T) for i in self._bubble_point.Psats]) 
            pcf = self._pcf(T, P, Psats)
            F_mol_vle = self._F_mol_vle
            mol_vle = self._mol_vle
            z = mol_vle / F_mol_vle
            v = F_mol_vle * solve_vle_vapor_mol_shgo(
                z, T, gamma.f, gamma.args, P, pcf * Psats, phi.f, phi.args, 
                dict(f_tol=self.y_tol, minimizer_kwargs=dict(f_tol=self.y_tol)),
            )
            self._z_last = z
        elif method == 'fixed-point':
            Psats = np.array([i(T) for i in
                              self._bubble_point.Psats])
            pcf_Psats_over_P = self._pcf(T, P, Psats) * Psats / P
            self._T = T
            if isinstance(self._phi, IdealFugacityCoefficients):
                y = self._y_iter(self._y, pcf_Psats_over_P, T, P)
            else:
                y = flx.aitken(self._y_iter, self._y, self.y_tol,
                               args=(pcf_Psats_over_P, T, P),
                               checkiter=False, 
                               checkconvergence=False, 
                               convergenceiter=5,
                               maxiter=self.maxiter)
            self._v = v = self._F_mol * self._V * y
            mask = v > self._mol_vle
            v[mask] = self._mol_vle[mask]
            v[v < 0.] = 0.
        else:
            raise RuntimeError(f"invalid method '{method}'")
        return v

class VLECache(Cache): load = VLE
del Cache, Equilibrium

@njit(cache=True)
def liquid_fugacity(mol_L, T, pcf_Psats, f_gamma, gamma_args):
    total_mol_L = mol_L.sum()
    if total_mol_L:
        x = mol_L / total_mol_L
        fugacity = x * f_gamma(x, T, *gamma_args) * pcf_Psats
    else:
        fugacity = np.ones_like(mol_L)
    return fugacity 

@njit(cache=True)
def vapor_fugacity(mol_v, T, P, f_phi, phi_args):
    total_mol_v = mol_v.sum()
    if total_mol_v:
        y = mol_v / total_mol_v
        fugacity = y * P * f_phi(y, T, P, *phi_args)
    else:
        fugacity = np.ones_like(mol_v)
    return fugacity

@njit(cache=True)
def gibbs_free_energy(mol, fugacity):
    fugacity[fugacity <= 0] = 1
    g_mix = (mol * np.log(fugacity)).sum()
    return g_mix

@njit(cache=True)
def vle_objective_function(mol_v, mol, T, f_gamma, gamma_args, P, pcf_Psats, f_phi, phi_args):
    mol_l = mol - mol_v
    g_mix_l = gibbs_free_energy(mol_l, liquid_fugacity(mol_l, T, pcf_Psats, f_gamma, gamma_args))
    g_mix_g = gibbs_free_energy(mol_v, vapor_fugacity(mol_v, T, P, f_phi, phi_args))
    g_mix = g_mix_l + g_mix_g
    return g_mix

def solve_vle_vapor_mol_shgo(
        mol, T, f_gamma, gamma_args, P, pcf_Psats, f_phi, phi_args, shgo_options,
    ):
    args = (mol, T, f_gamma, gamma_args, P, pcf_Psats, f_phi, phi_args)
    bounds = np.zeros([mol.size, 2])
    bounds[:, 1] = mol
    result = shgo(vle_objective_function, bounds, args, options=shgo_options)
    return result.x



# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
__all__ = ('chemical_units_of_measure', 
           'stream_units_of_measure',
           'ureg', 'get_dimensionality',
           'DisplayUnits', 
           'DisplayNotation',
           'AbsoluteUnitsOfMeasure', 
           'convert',
           'Quantity', 
           'parse_units_notation',
           'format_units', 
           'format_plot_units',
           'reformat_units')

from .exceptions import DimensionError

# %% Import unit registry

import pint
import os

# Set pint Unit Registry
appreg = pint.get_application_registry()
ureg = appreg.get()
ureg.default_format = '~P'
ureg._on_redefinition = 'warn' # Avoid breaking packages dependent on thermosteam (or biosteam)
if not getattr(pint, 'BioSTEAM_units_loaded', False): # Avoid reloading units of measure in pint if thermosteam module is reloaded
    ureg.load_definitions(os.path.dirname(os.path.realpath(__file__)) + '/units_of_measure.txt')
    
convert = ureg.convert
Quantity = ureg.Quantity
del os

# %% Functions

def parse_units_notation(value):
    if value is None:
        units = None
        notation = None
    elif ':' in value:
        units, notation = value.split(':')
        if not units: units = None
    else:
        units = value
        notation = None
    return units, notation

def format_degrees(units):
    r"""
    Format units of measure to have a latex degree symbol, if needed.

    Examples
    --------
    >>> format_degrees('degC')
    '^\\circ C'
    
    """
    if units.startswith('deg'):
        units = r'^\circ ' + units[3:]
    return units

def format_units_power(units, isnumerator=True, mathrm=True):
    r"""
    Format units of measure power sign to have a latex friendly format.

    Examples
    --------
    >>> format_units_power('m^3')
    '\\mathrm{m}^{3}'
    >>> format_units_power('m^3', isnumerator=False)
    '\\mathrm{m}^{-3}'
    
    """
    if '^' in units:
        units, power = units.split('^')
        units = format_degrees(units)
        if mathrm: units = '\mathrm{' + units + '}'
        power, *other = power.split(' ', 1)
        units += '^{' + (power if isnumerator else '-' + power) + '}'
        if other: units += other[0]
    else:
        if units != '%':
            units = format_degrees(units)      
            if mathrm: units = '\mathrm{' + units + '}'
        if not isnumerator:
            units = units + '^{-1}'
    return units

def format_units(units, ends='$', mathrm=True):
    r"""
    Format units of measure to have a latex friendly format.

    Examples
    --------
    >>> format_units('USD/m^3')
    '$\\mathrm{USD} \\cdot \\mathrm{m}^{-3}$'
    
    >>> format_units('USD/MT')
    '$\\mathrm{USD} \\cdot \\mathrm{MT}^{-1}$'
    
    """
    units = str(units)
    all_numerators = []
    all_denominators = []
    unprocessed_numerators, *unprocessed_denominators = units.split("/")
    all_numerators = unprocessed_numerators.split("*")
    for unprocessed_denominator in unprocessed_denominators:
        denominator, *unprocessed_numerators = unprocessed_denominator.split("*")
        all_numerators.extend(unprocessed_numerators)
        all_denominators.append(denominator)
    all_numerators = [format_units_power(i, True, mathrm) for i in all_numerators if i != '1']
    all_denominators = [format_units_power(i, False, mathrm) for i in all_denominators if i != '1']
    return ends + ' \cdot '.join(all_numerators + all_denominators).replace('$', '\$').replace('%', '\%') + ends

def reformat_units(name):
    left, right = name.split('[')
    units, right = right.split(']')
    return f"{left} [{format_units(units)}]{right}"

format_plot_units = format_units
get_dimensionality = ureg.get_dimensionality

# %% Manage conversion factors

class UnitsOfMeasure:
    __slots__ = ('units', 'units_container', 'dimensionality')
    
    def __bool__(self):
        return bool(self.units)
    
    def __str__(self):
        return self.units
    
    def __repr__(self):
        return f"{type(self).__name__}({repr(self.units)})"


class AbsoluteUnitsOfMeasure(UnitsOfMeasure):
    __slots__ = ('factor_cache',)
    _cache = {}
    
    def __new__(cls, units):
        if isinstance(units, cls):
            return units
        cache = cls._cache
        if units in cache:
            return cache[units]
        else:
            self = super().__new__(cls)
            self.units = units
            self.units_container = ureg[units]
            self.dimensionality = self.units_container.dimensionality
            self.factor_cache = {}
            cache[units] = self
            return self
    
    def conversion_factor(self, to_units):
        cache = self.factor_cache
        if to_units in cache:
            factor = cache[to_units]
        else:
            cache[to_units] = factor = ureg.convert(1., self.units_container, to_units)
        return factor
    
    def convert(self, value, to_units):
        return value * self.conversion_factor(to_units)
    
    def unconvert(self, value, from_units):
        return value / self.conversion_factor(from_units)


class RelativeUnitsOfMeasure(UnitsOfMeasure):
    __slots__ = ()
    _cache = {}
    
    def __new__(cls, units):
        if isinstance(units, cls):
            return units
        cache = cls._cache
        if units in cache:
            return cache[units]
        else:
            self = super().__new__(cls)
            self.units = units
            self.units_container = ureg.Unit(units)
            self.dimensionality = self.units_container.dimensionality
            cache[units] = self
            return self
    
    def conversion_factor(self, to_units):
        return ureg.convert(1., self.units_container, to_units)
    
    def convert(self, value, to_units):
        return ureg.convert(value, self.units_container, to_units)
    
    def unconvert(self, value, from_units):
        return ureg.convert(value, from_units, self.units_container)


# %% Manage display units

class DisplayNotation:
    """
    Create a DisplayNotation object where default units for representation are stored.
    
    Examples
    --------
    Its possible to change the default units of measure for the Stream show method:
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water'], cache=True)
    >>> tmo.Stream.display_notation.flow = '.2g'
    >>> stream = tmo.Stream('stream', Water=1.324, units='kg/hr')
    >>> stream.show()
    Stream: stream
    phase: 'l', T: 298.15 K, P: 101325 Pa
    flow (kmol/hr): Water  0.073
    
    >>> # Change back to kmol/hr
    >>> tmo.Stream.display_notation.flow = '.3g'
    
    """
    __slots__ = ('T', 'P', 'flow')
    
    def __init__(self, T, P, flow):
        self.T = T
        self.P = P
        self.flow = flow

    def __repr__(self):
        sig = ', '.join([f"{i}={getattr(self, i)!r}'" for i in self.__slots__])
        return f'{type(self).__name__}({sig})'


class DisplayUnits:
    """
    Create a DisplayUnits object where default units for representation are stored.
    
    Examples
    --------
    Its possible to change the default units of measure for the Stream show method:
        
    >>> import thermosteam as tmo
    >>> tmo.settings.set_thermo(['Water'], cache=True)
    >>> tmo.Stream.display_units.flow = 'kg/hr'
    >>> stream = tmo.Stream('stream', Water=1, units='kg/hr')
    >>> stream.show()
    Stream: stream
    phase: 'l', T: 298.15 K, P: 101325 Pa
    flow (kg/hr): Water  1
    
    >>> # Change back to kmol/hr
    >>> tmo.Stream.display_units.flow = 'kmol/hr'
    
    """
    def __init__(self, **display_units):
        dct = self.__dict__
        dct.update(display_units)
        dct['dims'] = {}
        list_keys = []
        for k, v in display_units.items():
            try: # Assume units is one string
                dims = ureg[v].dimensionality
            except:
                try: # Assume units are a list of possible units
                    dims = [ureg[i].dimensionality for i in v]
                    list_keys.append(k)
                except: # Assume the user uses value as an option, and ignores units
                    dims = v
            self.dims[k] = dims
        for k in list_keys:
            dct[k] = dct[k][0] # Default units is first in list
    
    def __setattr__(self, name, unit):
        if name not in self.__dict__:
            raise AttributeError(f"can't set display units for '{name}'")
        if isinstance(unit, str):
            name_dim = self.dims[name]
            unit_dim = getattr(ureg, unit).dimensionality
            if isinstance(name_dim, list):
                if unit_dim not in name_dim:
                    name_dim = [f"({i})" for i in name_dim]
                    raise DimensionError(f"dimensions for '{name}' must be either {', '.join(name_dim[:-1])} or {name_dim[-1]}; not ({unit_dim})")    
            else:
                if name_dim != unit_dim:
                    raise DimensionError(f"dimensions for '{name}' must be in ({name_dim}), not ({unit_dim})")
        object.__setattr__(self, name, unit)
    
    def __repr__(self):
        sig = ', '.join([f"{i}={j!r}" for i,j in self.__dict__.items() if i != 'dims'])
        return f'{type(self).__name__}({sig})'


# %% Units of measure

chemical_units_of_measure = {
    'MW': AbsoluteUnitsOfMeasure('g/mol'),
    'T': RelativeUnitsOfMeasure('K'),
    'Tr': RelativeUnitsOfMeasure('K'),
    'Tm': RelativeUnitsOfMeasure('K'),
    'Tb': RelativeUnitsOfMeasure('K'),
    'Tbr': RelativeUnitsOfMeasure('K'),
    'Tt': RelativeUnitsOfMeasure('K'),
    'Tc': RelativeUnitsOfMeasure('K'),
    'P': AbsoluteUnitsOfMeasure('Pa'),
    'Pr': AbsoluteUnitsOfMeasure('Pa'),
    'Pc': AbsoluteUnitsOfMeasure('Pa'),
    'Psat': AbsoluteUnitsOfMeasure('Pa'),
    'Pt': AbsoluteUnitsOfMeasure('Pa'),
    'V': AbsoluteUnitsOfMeasure('m^3/mol'),
    'Vc': AbsoluteUnitsOfMeasure('m^3/mol'),
    'Cp': AbsoluteUnitsOfMeasure('J/g/K'),
    'Cn': AbsoluteUnitsOfMeasure('J/mol/K'),
    'R': AbsoluteUnitsOfMeasure('J/mol/K'),
    'rho': AbsoluteUnitsOfMeasure('kg/m^3'),
    'rhoc': AbsoluteUnitsOfMeasure('kg/m^3'),
    'nu': AbsoluteUnitsOfMeasure('m^2/s'),
    'alpha': AbsoluteUnitsOfMeasure('m^2/s'),
    'mu': AbsoluteUnitsOfMeasure('Pa*s'),
    'sigma': AbsoluteUnitsOfMeasure('N/m'),
    'kappa': AbsoluteUnitsOfMeasure('W/m/K'),
    'Hvap': AbsoluteUnitsOfMeasure('J/mol'),
    'H': AbsoluteUnitsOfMeasure('J/mol'),  
    'Hf': AbsoluteUnitsOfMeasure('J/mol'), 
    'Hc': AbsoluteUnitsOfMeasure('J/mol'), 
    'Hfus': AbsoluteUnitsOfMeasure('J/mol'), 
    'Hsub': AbsoluteUnitsOfMeasure('J/mol'),
    'HHV': AbsoluteUnitsOfMeasure('J/mol'),
    'LHV': AbsoluteUnitsOfMeasure('J/mol'),
    'S': AbsoluteUnitsOfMeasure('J/K/mol'),
    'S0': AbsoluteUnitsOfMeasure('J/K/mol'),
    'G': AbsoluteUnitsOfMeasure('J/mol'), 
    'U': AbsoluteUnitsOfMeasure('J/mol'),
    'H_excess': AbsoluteUnitsOfMeasure('J/mol'), 
    'S_excess': AbsoluteUnitsOfMeasure('J/mol'),
    'dipole': AbsoluteUnitsOfMeasure('Debye'),
    'delta': AbsoluteUnitsOfMeasure('Pa^0.5'),
    'epsilon': AbsoluteUnitsOfMeasure(''),
}
stream_units_of_measure = {
    'mol': AbsoluteUnitsOfMeasure('kmol/hr'),
    'mass': AbsoluteUnitsOfMeasure('kg/hr'),
    'vol': AbsoluteUnitsOfMeasure('m^3/hr'),
    'F_mass': AbsoluteUnitsOfMeasure('kg/hr'),
    'F_mol': AbsoluteUnitsOfMeasure('kmol/hr'),
    'F_vol': AbsoluteUnitsOfMeasure('m^3/hr'),
    'cost': AbsoluteUnitsOfMeasure('USD/hr'),
    'HHV': AbsoluteUnitsOfMeasure('kJ/hr'),
    'LHV': AbsoluteUnitsOfMeasure('kJ/hr'),
    'Hvap': AbsoluteUnitsOfMeasure('kJ/hr'),
    'Hf': AbsoluteUnitsOfMeasure('kJ/hr'), 
    'S0': AbsoluteUnitsOfMeasure('kJ/K/hr'), 
    'Hc': AbsoluteUnitsOfMeasure('kJ/hr'), 
    'H': AbsoluteUnitsOfMeasure('kJ/hr'),
    'S': AbsoluteUnitsOfMeasure('kJ/K/hr'),
    'G': AbsoluteUnitsOfMeasure('kJ/hr'),
    'U': AbsoluteUnitsOfMeasure('kJ/hr'),
    'C': AbsoluteUnitsOfMeasure('kJ/hr/K'),
}
for i in ('T', 'P', 'mu', 'V', 'rho', 'sigma',
          'kappa', 'nu', 'epsilon', 'delta',
          'Psat', 'Cp', 'Cn', 'alpha'):
    stream_units_of_measure[i] = chemical_units_of_measure[i]

power_utility_units_of_measure = {
    'cost': AbsoluteUnitsOfMeasure('USD/hr'),
    'rate': AbsoluteUnitsOfMeasure('kW'),
    'consumption': AbsoluteUnitsOfMeasure('kW'),
    'production': AbsoluteUnitsOfMeasure('kW'),
}

heat_utility_units_of_measure = {
    'cost': AbsoluteUnitsOfMeasure('USD/hr'),
    'flow': AbsoluteUnitsOfMeasure('kmol/hr'),
    'duty': AbsoluteUnitsOfMeasure('kJ/hr'),
}

definitions = {
    'MW': 'Molecular weight',
    'T': 'Temperature',
    'Tr': 'Reduced temperature',
    'Tm': 'Melting point temperature',
    'Tb': 'Boiling point temperature',
    'Tbr': 'Reduced boiling point temperature',
    'Tt': 'Triple point temperature',
    'Tc': 'Critical point temperature',
    'P': 'Pressure',
    'Pr': 'Reduced pressure',
    'Pc': 'Critical point pressure',
    'Psat': 'Saturated vapor pressure',
    'Pt': 'Triple point pressure',
    'V': 'Molar volume',
    'Vc': 'Critical point volume',
    'Cp': 'Specific heat capacity',
    'Cn': 'Molar heat capacity',
    'rho': 'Density',
    'rhoc': 'Critical point density',
    'nu': 'Kinematic viscosity',
    'mu': 'hydraulic viscosity',
    'sigma': 'Surface tension',
    'kappa': 'Thermal conductivity',
    'alpha': 'Thermal diffusivity',
    'Hvap': 'Heat of vaporization',
    'H': 'Enthalpy',
    'Hf': 'Heat of formation',
    'Hc': 'Heat of combustion', 
    'Hfus': 'Heat of fusion',
    'Hsub': 'Heat of sublimation',
    'S': 'Entropy',
    'S0': 'Absolute entropy of formation',
    'G': 'Gibbs free energy',
    'U': 'Internal energy',
    'H_excess': 'Excess enthalpy',
    'S_excess': 'Excess entropy',
    'R': 'Universal gas constant',
    'Zc': 'Critical compressibility',
    'dZ': 'Change in compressibility factor',
    'omega': 'Acentric factor',
    'dipole': 'Dipole momment',
    'delta': 'Solubility parameter',
    'epsilon': 'Relative permittivity',
    'similarity_variable': 'Heat capacity similarity variable',
    'iscyclic_aliphatic': 'Whether a chemical is cyclic aliphatic',
    'has_hydroxyl': 'Whether a polar chemical has hydroxyl groups',
    'atoms': 'Atom-count pairs'
}

types = {'atoms': 'Dict[str, int]'}
types['iscyclic_aliphatic'] = types['has_hydroxy'] = 'bool'

# Synonyms
definitions['ω'] = definitions['omega']

# Phase properties
for var in ('mu', 'Cn', 'H', 'S', 'V', 'kappa', 'H_excess', 'S_excess'):
    definition = definitions[var].lower()
    for tag, phase in zip(('s', 'l', 'g'), ('Solid ', 'Liquid ', 'Gas ')):
        phase_var = var + '.' + tag
        phase_var2 = var + '_' + tag
        definitions[phase_var] = definitions[phase_var2] = phase + definition

pint.BioSTEAM_units_loaded = True


# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
from ..exceptions import UndefinedChemical
from ..base import SparseVector, sparse_vector

__all__ = ('get_stoichiometric_array',
           'get_stoichiometric_string',
           'str2dct', 'dct2str', 'arr2dct')

def get_stoichiometric_array(reaction, chemicals):
    """Return stoichiometric array given a string defining the reaction and chemicals."""
    isa = isinstance
    if isa(reaction, dict):
        stoichiometry_dict = {i: float(j) for i, j in reaction.items() if j}
    elif isa(reaction, str):
        stoichiometry_dict = str2dct(reaction)
    elif hasattr(reaction, '__iter__'):
        return sparse_vector(reaction, chemicals.size)
    else:
        raise ValueError(f"reaction must be either a str, dict, or array; not a '{type(reaction).__name__}' object")
    stoichiometric_array = dct2arr(stoichiometry_dict, chemicals)
    return stoichiometric_array

def get_stoichiometric_string(reaction, chemicals):
    """Return a string defining the reaction given the stoichiometric array and chemicals."""
    if isinstance(reaction, dict):
        stoichiometric_dict = reaction
    elif hasattr(reaction, '__iter__'):
        stoichiometric_dict = arr2dct(reaction, chemicals)
    else:
        raise ValueError(f"reaction must be either a str or an array; not a '{type(reaction).__name__}' object")
    return dct2str(stoichiometric_dict)

def dct2arr(dct, chemicals):
    idct = {} # same as dct but using integers as keys
    chemical_index = chemicals._index
    chemical_groups = chemicals.chemical_groups
    for ID, coefficient in dct.items():
        if ID in chemical_groups: 
            raise ValueError(
                f"'{ID}' is a chemical group; chemical groups cannot be used "
                 "in reaction definition"
            )
        if ID not in chemical_index:
            raise UndefinedChemical(ID)
        idct[chemical_index[ID]] = coefficient
    return SparseVector.from_dict(idct, chemicals.size)

def split_coefficient(nID, sign):
    for i, letter in enumerate(nID):
        if letter != 'e' and (letter.isalpha() or letter in '()[]{}'): break
    if i: 
        ID = nID[i:]
        n = sign * float(nID[:i])
    else: 
        ID = nID
        n = sign
    return n, ID

def extract_coefficients(nIDs, dct, sign):
    for nID in nIDs:
        n, ID = split_coefficient(nID, sign)
        if ID in dct:
            raise ValueError('chemicals can only appear once in a reaction; '
                            f'multiple instances of {repr(ID)} found')
        dct[ID] = n

def str2dct(reaction) -> dict:
    reaction = reaction.replace(' ', '')
    left, right = reaction.split('->')
    reactants = left.split('+')
    products = right.split('+')
    dct = {}
    extract_coefficients(reactants, dct, -1.)
    extract_coefficients(products, dct, 1.)
    return dct

def dct2str(dct):
    if not dct: return "no reaction"
    left = []
    right = []
    for ID, n in dct.items():
        nf = format(n, '.3g')
        if nf == '-1': left.append(ID)
        elif nf == '1': right.append(ID)
        elif n < 0: left.append(f"{-n:.3g} {ID}")
        else: right.append(f"{nf} {ID}")
    left = ' + '.join(left)
    right = ' + '.join(right)
    reaction = left + ' -> ' + right
    return reaction

def arr2dct(arr, chemicals):
    IDs = chemicals.IDs
    if hasattr(arr, 'dct'):
        dct = arr.dct
        index = sorted(dct.keys())
        return {IDs[i]: dct[i] for i in index}
    else:
        return {IDs[i]: j for i, j in enumerate(arr) if j}




# -*- coding: utf-8 -*-
# BioSTEAM: The Biorefinery Simulation and Techno-Economic Analysis Modules
# Copyright (C) 2020-2023, Yoel Cortes-Pena <yoelcortes@gmail.com>
# 
# This module is under the UIUC open-source license. See 
# github.com/BioSTEAMDevelopmentGroup/biosteam/blob/master/LICENSE.txt
# for license details.
"""
"""
from .functor import functor_lookalike
from ..utils import read_only, cucumber

__all__ = ('PhaseHandle',
           'MockPhaseHandle',
           'MockPhaseTHandle',
           'MockPhaseTPHandle',
           'PhaseTHandle', 
           'PhaseTPHandle',
           'PhaseFunctorBuilder',
           'PhaseTFunctorBuilder',
           'PhaseTPFunctorBuilder')


# %% Abstract class    

@cucumber
@read_only
@functor_lookalike
class PhaseHandle:
    __slots__ = ('Tc', 'var', 's', 'l', 'g')
    
    def __init__(self, var, s, l, g, Tc=None):
        setattr = object.__setattr__
        setattr(self, 'var', var)
        setattr(self, 's', s)
        setattr(self, 'l', l)
        setattr(self, 'g', g)
        setattr(self, 'Tc', Tc)
    
    @property
    def S(self): return self.s
    @property
    def L(self): return self.l
    
    def __iter__(self):
        return iter((('s', self.s), ('l', self.l), ('g', self.g)))
    
    def __bool__(self):
        return any((self.s, self.l, self.g)) 
    
    def copy(self):
        return self.__class__(
            self.var, self.s.copy(), self.l.copy(), self.g.copy(), self.Tc,
        )
    __copy__ = copy
    
    def show(self):
        print(self)
    _ipython_display_ = show

@cucumber
@read_only
@functor_lookalike
class MockPhaseHandle:
    __slots__ = ('var', 'model')
    
    def __init__(self, var, model):
        setattr = object.__setattr__
        setattr(self, 'var', var)
        setattr(self, 'model', model)
    
    @property
    def S(self): return self.model
    @property
    def L(self): return self.model
    @property
    def s(self): return self.model
    @property
    def l(self): return self.model
    @property
    def g(self): return self.model
    
    def __iter__(self):
        return iter((('s', self.model), ('l', self.model), ('g', self.model)))
    
    def __bool__(self):
        return bool(self.model)
    
    def copy(self):
        return self.__class__(
            self.var, self.model.copy(), 
        )
    __copy__ = copy
    
    def show(self):
        print(self)
    _ipython_display_ = show
    
    
# %% Pure component

class PhaseTHandle(PhaseHandle):
    __slots__ = ()
    force_gas_critical_phase = False
    
    def __call__(self, phase, T, P=None):
        if self.force_gas_critical_phase and T > self.Tc: phase = 'g'
        return getattr(self, phase)(T)
    
    
class PhaseTPHandle(PhaseHandle):
    __slots__ = ()
    force_gas_critical_phase = False
    
    def __call__(self, phase, T, P=None):
        if self.force_gas_critical_phase and T > self.Tc: phase = 'g'
        return getattr(self, phase)(T, P)


class MockPhaseTHandle(MockPhaseHandle):
    __slots__ = ()
    
    def __call__(self, phase, T, P=None):
        return self.model(T)
    
    
class MockPhaseTPHandle(MockPhaseHandle):
    __slots__ = ()
    
    def __call__(self, phase, T, P=None):
        return self.model(T, P)

# %% Builders

class PhaseFunctorBuilder:
    __slots__ = ('var', 's', 'l', 'g', 'build_functors')
    
    def __init__(self, var, s, l, g):
        self.var = var
        self.s = s
        self.l = l
        self.g = g
        
    def __call__(self, sdata, ldata, gdata, Tc):
        phases = ('s', 'g', 'l')
        builders = (self.s, self.g, self.l)
        phases_data = (sdata, gdata, ldata)
        slg = {}
        for phase, builder, data in zip(phases, builders, phases_data):
            if builder:
                functor = builder.from_args(data)
                slg[phase] = functor
        return self.PhaseHandle(self.var, **slg, Tc=Tc)


class PhaseTFunctorBuilder(PhaseFunctorBuilder):
    __slots__ = ()
    PhaseHandle = PhaseTHandle
    
        
class PhaseTPFunctorBuilder(PhaseFunctorBuilder):
    __slots__ = ()
    PhaseHandle = PhaseTPHandle
    
    